from __future__ import annotations
import abc
from typing import Dict, List, Set, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction, Ratings
    from clayrs.recsys.graphs.graph import UserNode, Node, Graph, BipartiteDiGraph
    from clayrs.recsys.methodology import Methodology

from clayrs.recsys.graphs.graph import ItemNode
from clayrs.recsys.methodology import TestRatingsMethodology
from clayrs.recsys.algorithm import Algorithm


class GraphBasedAlgorithm(Algorithm):
    """
    Abstract class for the graph-based algorithms
    """

    def __init__(self):
        # FUTURE WORK: this can be expanded in making the page rank keeping also PropertyNodes, etc.
        self._nodes_to_keep = {ItemNode}

    def filter_result(self, graph: BipartiteDiGraph, result: Dict, filter_list: Union[List[Node], None],
                      user_node: UserNode) -> Dict:
        """
        Method which filters and cleans the result dict based on the parameters passed

        If `filter_list` parameter is not None, then the final dict will contains items and score only for items
        in said `filter_list`

        Otherwise if no filter list is specified, then all unrated items by the user will be returned in the final
        dict

        Args:
            graph: Directed graph which models interactions between users and items
            result: dictionary representing the result (keys are nodes and values are their score)
            filter_list: list of the items for which a ranking score/score prediction must be computed.
                If None all unrated items for the user will be ranked/score predicted.
            user_node: Node of the particular user for which we want to compute rank/score prediction
        """

        def must_keep(node: object, user_profile):
            must_be_kept = True
            if node in user_profile or type(node) not in self._nodes_to_keep:
                must_be_kept = False

            return must_be_kept

        if filter_list is not None:
            filtered_keys = result.keys() & set(filter_list)
            filtered_result = {k: result[k] for k in filtered_keys}
        else:
            extracted_profile = set(graph.get_successors(user_node))
            filtered_result = {k: result[k] for k in result.keys() if must_keep(k, extracted_profile)}

        return filtered_result

    @abc.abstractmethod
    def predict(self, all_users: Set[str], graph: Graph, test_set: Ratings,
                methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 1) -> List[Interaction]:
        """
        Abstract method that predicts how much a user will like unrated items.
        If the algorithm is not a PredictionScore Algorithm, implement this method like this:

        ```python
        def predict():
            raise NotPredictionAlg
        ```

        One can specify on which items score prediction must be performed for each user with the `filter_dict`
        parameter, in this case every user is mapped with a list of items for which a prediction score must be computed.
        Otherwise, for **ALL** unrated items a prediction score will be computed for each user.

        Args:
            all_users: Set of user id for which a recommendation list must be generated
            graph: A graph previously instantiated
            test_set: Ratings object which represents the ground truth of the split considered
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Returns:
            List of Interactions object where the 'score' attribute is the rating predicted by the algorithm
        """
        raise NotImplementedError

    @abc.abstractmethod
    def rank(self, all_users: Set[str], graph: Graph, test_set: Ratings,
             recs_number: int = None, methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 1) -> List[Interaction]:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        ```python
        def rank():
            raise NotRankingAlg
        ```

        One can specify which items must be ranked for each user with the `filter_dict` parameter,
        in this case every user is mapped with a list of items for which a ranking score must be computed.
        Otherwise, **ALL** unrated items will be ranked for each user.

        Args:
            all_users: Set of user id for which a recommendation list must be generated
            graph: A graph previously instantiated
            test_set: Ratings object which represents the ground truth of the split considered
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
                a single user
        """
        raise NotImplementedError



import pandas as pd
from unittest import TestCase
import os

from clayrs.content_analyzer import Ratings
from clayrs.recsys import TopKDegreeCentrality
from clayrs.recsys.graphs.feature_selection import TopKPageRank
from clayrs.recsys.graphs.feature_selection import feature_selector
from clayrs.recsys.graphs.nx_implementation.nx_full_graphs import NXFullGraph
from test import dir_test_files


class TestFeatureSelection(TestCase):

    def setUp(self) -> None:
        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')
        user_dir = os.path.join(dir_test_files, 'complex_contents', 'users_codified/')

        df = pd.DataFrame.from_dict({'user_id': ["1", "1", "2", "2", "2", "3", "4", "4"],
                                     'item_id': ["tt0112281", "tt0113041", "tt0112453", "tt0112346",
                                                 "tt0112302", "tt0113101", "tt0113277", "tt0113228"],
                                     'score': [5, 3, 1, 5, 4, 2, 2, 4]})
        rat = Ratings.from_dataframe(df)

        # all properties from the dbpedia item repr extracted
        self.g: NXFullGraph = NXFullGraph(rat,
                                          user_contents_dir=user_dir,
                                          item_contents_dir=movies_dir,
                                          item_exo_properties={'dbpedia'},
                                          user_exo_properties={'local'}
                                          )

        self.users_target_nodes = list(self.g.user_nodes)
        self.items_target_nodes = list(self.g.item_nodes)

    def test_feature_selection_both(self):

        user_alg = TopKPageRank(k=1)
        item_alg = TopKDegreeCentrality(k=2)

        graph_fs = feature_selector(self.g, fs_algorithm_user=user_alg, fs_algorithm_item=item_alg,
                                    user_target_nodes=self.users_target_nodes,
                                    item_target_nodes=self.items_target_nodes)

        self.assertNotEqual(graph_fs, self.g)

        # we manually perform feature selection in order to get all labels that should not be present in the graph
        # after feature selection is performed
        labels_to_remove = user_alg.perform(self.g, self.users_target_nodes, mode='to_remove')
        labels_to_remove.extend(item_alg.perform(self.g, self.items_target_nodes, mode='to_remove'))

        # here we check that labels less important found at the previous step are not present in the graph
        # after feature selection
        for property_node in graph_fs.property_nodes:
            predecessors = graph_fs.get_predecessors(property_node)
            for pred in predecessors:
                link_label = graph_fs.get_link_data(pred, property_node).get('label')
                self.assertNotIn(link_label, labels_to_remove)

        # test inplace feature selection
        graph_fs = feature_selector(self.g, fs_algorithm_user=user_alg, fs_algorithm_item=item_alg,
                                    user_target_nodes=self.users_target_nodes,
                                    item_target_nodes=self.items_target_nodes, inplace=True)

        self.assertEqual(graph_fs, self.g)

    def test_feature_selection_only_users(self):

        user_alg = TopKDegreeCentrality(k=3)

        graph_fs = feature_selector(self.g, fs_algorithm_user=user_alg, user_target_nodes=self.users_target_nodes)

        # we manually perform feature selection in order to get all labels that should not be present in the graph
        # after feature selection is performed
        labels_to_remove = user_alg.perform(self.g, self.users_target_nodes, mode='to_remove')

        # here we check that labels less important found at the previous step are not present in the graph
        # after feature selection
        for property_node in graph_fs.property_nodes:
            predecessors = graph_fs.get_predecessors(property_node)
            for pred in predecessors:
                link_label = graph_fs.get_link_data(pred, property_node).get('label')
                self.assertNotIn(link_label, labels_to_remove)

    def test_feature_selection_only_items(self):

        item_alg = TopKPageRank(k=3)

        graph_fs = feature_selector(self.g, fs_algorithm_item=item_alg, item_target_nodes=self.items_target_nodes)

        # we manually perform feature selection in order to get all labels that should not be present in the graph
        # after feature selection is performed
        labels_to_remove = item_alg.perform(self.g, self.items_target_nodes, mode='to_remove')

        # here we check that labels less important found at the previous step are not present in the graph
        # after feature selection
        for property_node in graph_fs.property_nodes:
            predecessors = graph_fs.get_predecessors(property_node)
            for pred in predecessors:
                link_label = graph_fs.get_link_data(pred, property_node).get('label')
                self.assertNotIn(link_label, labels_to_remove)

    def test_special_cases_feature_selection(self):
        # algorithm failed to converge both for items and users
        graph_fs = feature_selector(self.g, fs_algorithm_user=TopKPageRank(max_iter=0),
                                    fs_algorithm_item=TopKPageRank(max_iter=0))

        # since both fs algorithm failed, the graph remained untouched
        self.assertEqual(graph_fs, self.g)

        # check that the original graph is returned if no fs algorithm is defined
        graph_fs = feature_selector(self.g, user_target_nodes=self.users_target_nodes,
                                    item_target_nodes=self.items_target_nodes)

        self.assertEqual(graph_fs, self.g)

        # check that when no target list is defined, all users nodes and all items nodes are used as targets
        graph_fs_no_target = feature_selector(self.g, fs_algorithm_user=TopKPageRank(k=2),
                                              fs_algorithm_item=TopKPageRank(k=2))

        graph_fs_w_target = feature_selector(self.g, fs_algorithm_user=TopKPageRank(k=2),
                                             fs_algorithm_item=TopKPageRank(k=2),
                                             user_target_nodes=self.users_target_nodes,
                                             item_target_nodes=self.items_target_nodes)

        self.assertEqual(graph_fs_no_target, graph_fs_w_target)



from unittest import TestCase
import numpy as np

from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
    Centroid, Sum, SingleToken


class TestCentroid(TestCase):
    def test_combine(self):
        z = np.ndarray(shape=(3, 3))

        z[0, :] = [1, 1, 1]
        z[1, :] = [2, 2, 2]
        z[2, :] = [3, 3, 3]

        combiner = Centroid()
        result = combiner.combine(z)

        expected = np.ndarray(shape=(3, ))
        expected[:] = [2, 2, 2]

        self.assertTrue((result == expected).all())


class TestSum(TestCase):
    def test_combine(self):
        z = np.ndarray(shape=(3, 3))

        z[0, :] = [1, 9, 1]
        z[1, :] = [7, 2, 4]
        z[2, :] = [3, 5, 3]

        combiner = Sum()
        result = combiner.combine(z)

        expected = np.ndarray(shape=(3, ))
        expected[:] = [11, 16, 8]

        self.assertTrue((result == expected).all())


class TestSingleToken(TestCase):
    def test_combine(self):
        z = np.ndarray(shape=(3, 3))

        z[0, :] = [1, 9, 1]
        z[1, :] = [7, 2, 4]
        z[2, :] = [3, 5, 3]

        combiner = SingleToken(0)
        result = combiner.combine(z)

        expected = np.ndarray(shape=(3, ))
        expected[:] = [1, 9, 1]

        self.assertTrue((result == expected).all())

        combiner = SingleToken(2)
        result = combiner.combine(z)

        expected = np.ndarray(shape=(3, ))
        expected[:] = [3, 5, 3]

        self.assertTrue((result == expected).all())

    def test_raise(self):
        z = np.ndarray(shape=(3, 3))

        z[0, :] = [1, 9, 1]
        z[1, :] = [7, 2, 4]
        z[2, :] = [3, 5, 3]

        with self.assertRaises(IndexError):
            SingleToken(99).combine(z)



import os.path
import shutil
import unittest

from clayrs.content_analyzer import CSVFile, Ratings
from clayrs.evaluation import Precision, Recall
from clayrs.recsys import HoldOutPartitioning, CentroidVector, CosineSimilarity, SkKNN, ClassifierRecommender, \
    LinearPredictor, SkLinearRegression, SkSVC, KFoldPartitioning, NXPageRank
from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg
from clayrs.recsys.experiment import ContentBasedExperiment, GraphBasedExperiment
from test import dir_test_files

rat_path = os.path.join(dir_test_files, "new_ratings.csv")
items_dir = os.path.join(dir_test_files, "complex_contents", "movies_codified")

# for coverage
num_cpus = 1


class TestExperiment(unittest.TestCase):
    def test_already_existent_output_folder(self):
        rat = Ratings(CSVFile(rat_path))

        # test a subclass since Experiment is abstract
        cbe = ContentBasedExperiment(rat,
                                     partitioning_technique=HoldOutPartitioning(),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity(),
                                                                    threshold=2)],
                                     items_directory=items_dir,
                                     output_folder="my_experiment")

        cbe.rank(num_cpus=num_cpus)

        self.assertTrue(os.path.isdir("my_experiment"))

        with self.assertRaises(FileExistsError):
            # same output folder and overwrite parameter not set
            cbe = ContentBasedExperiment(rat,
                                         partitioning_technique=HoldOutPartitioning(),
                                         algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity(),
                                                                        threshold=2)],
                                         items_directory=items_dir,
                                         output_folder="my_experiment")

            cbe.rank(num_cpus=num_cpus)

        # same output folder and overwrite parameter is set
        cbe = ContentBasedExperiment(rat,
                                     partitioning_technique=HoldOutPartitioning(),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity(),
                                                                    threshold=2)],
                                     items_directory=items_dir,
                                     output_folder="my_experiment",
                                     overwrite_if_exists=True,
                                     report=True)

        cbe.rank(num_cpus=num_cpus)

        self.assertTrue(os.path.isdir("my_experiment"))

        # different output folder
        cbe = ContentBasedExperiment(rat,
                                     partitioning_technique=HoldOutPartitioning(),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity(),
                                                                    threshold=2)],
                                     items_directory=items_dir,
                                     output_folder="my_experiment_2",
                                     overwrite_if_exists=True,
                                     report=True)

        cbe.rank(num_cpus=num_cpus)

        self.assertTrue(os.path.isdir("my_experiment"))
        self.assertTrue(os.path.isdir("my_experiment_2"))

        # clean folders created
        shutil.rmtree("my_experiment", ignore_errors=False)
        shutil.rmtree("my_experiment_2", ignore_errors=False)

    def test_eval_metrics(self):
        rat = Ratings(CSVFile(rat_path))

        # test a subclass since Experiment is abstract
        cbe = ContentBasedExperiment(rat,
                                     partitioning_technique=HoldOutPartitioning(),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     LinearPredictor({'Plot': 0}, regressor=SkLinearRegression())],
                                     items_directory=items_dir,
                                     output_folder="my_experiment",
                                     metric_list=[Precision(), Recall()])

        cbe.predict(num_cpus=num_cpus)

        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "eva_sys_results.csv")))
        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "eva_users_results.csv")))

        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "eva_sys_results.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "eva_users_results.csv")))

    def test_eval_metrics_with_report(self):
        rat = Ratings(CSVFile(rat_path))

        # test a subclass since Experiment is abstract
        cbe = ContentBasedExperiment(rat,
                                     partitioning_technique=HoldOutPartitioning(),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     LinearPredictor({'Plot': 0}, regressor=SkLinearRegression())],
                                     items_directory=items_dir,
                                     output_folder="my_experiment",
                                     metric_list=[Precision(), Recall()],
                                     report=True)

        cbe.predict(num_cpus=num_cpus)

        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "eva_sys_results.csv")))
        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "eva_users_results.csv")))
        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "eva_report.yml")))

        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "eva_sys_results.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "eva_users_results.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "eva_report.yml")))

    def doCleanups(self) -> None:
        shutil.rmtree("my_experiment", ignore_errors=True)


class TestContentBasedExperiment(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        rat = Ratings(CSVFile(rat_path))
        cls.rat = rat

        cls.cbe = ContentBasedExperiment(rat,
                                         partitioning_technique=KFoldPartitioning(random_state=42, n_splits=2),
                                         algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                         ClassifierRecommender({'Plot': 0}, classifier=SkSVC()),
                                                         ClassifierRecommender({'Plot': 0}, classifier=SkKNN(2)),
                                                         LinearPredictor({'Plot': 0}, regressor=SkLinearRegression())],
                                         items_directory=items_dir,
                                         output_folder="my_experiment")

    def test_rank(self):
        self.cbe.rank(num_cpus=num_cpus)

        # check dirs have been created
        self.assertTrue(os.path.isdir("my_experiment"))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "CentroidVector_1")))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "ClassifierRecommender_1")))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "ClassifierRecommender_2")))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "LinearPredictor_1")))

        # check train test splits have been created
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split1.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split1.csv")))

        # check rank created for centroid vector 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "rs_rank_split1.csv")))

        # check rank created for classifier recommender 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_1", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_1", "rs_rank_split1.csv")))

        # check rank created for classifier recommender 2
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_2", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_2", "rs_rank_split1.csv")))

        # check rank created for linear predictor 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_rank_split1.csv")))

    def test_rank_with_report(self):
        cbe = ContentBasedExperiment(self.rat,
                                     partitioning_technique=KFoldPartitioning(random_state=42, n_splits=2),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkSVC()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkKNN(2)),
                                                     LinearPredictor({'Plot': 0}, regressor=SkLinearRegression())],
                                     items_directory=items_dir,
                                     output_folder="my_experiment",
                                     report=True)

        cbe.rank(num_cpus=num_cpus)

        # check report created for centroid vector 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "CentroidVector_1", "rs_report.yml")))

        # check report created for classifier recommender 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_1", "rs_report.yml")))

        # check report created for classifier recommender 2
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "ClassifierRecommender_2", "rs_report.yml")))

        # check report created for linear predictor 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_report.yml")))

    def test_predict(self):
        self.cbe.predict(num_cpus=num_cpus)

        # check dirs have been created
        self.assertTrue(os.path.isdir("my_experiment"))
        self.assertFalse(os.path.isdir(os.path.join("my_experiment", "CentroidVector_1")))
        self.assertFalse(os.path.isdir(os.path.join("my_experiment", "ClassifierRecommender_1")))
        self.assertFalse(os.path.isdir(os.path.join("my_experiment", "ClassifierRecommender_2")))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "LinearPredictor_1")))

        # check train test splits have been created
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split1.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split1.csv")))

        # check predict created for linear predictor
        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_rank_split0.csv")))
        self.assertFalse(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_rank_split1.csv")))

    def test_predict_with_report(self):
        cbe = ContentBasedExperiment(self.rat,
                                     partitioning_technique=KFoldPartitioning(random_state=42, n_splits=2),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkSVC()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkKNN(2)),
                                                     LinearPredictor({'Plot': 0}, regressor=SkLinearRegression())],
                                     items_directory=items_dir,
                                     output_folder="my_experiment",
                                     report=True)

        cbe.predict(num_cpus=num_cpus)

        # check report created for linear predictor 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "LinearPredictor_1", "rs_report.yml")))

    def test_no_one_predictor(self):
        cbe = ContentBasedExperiment(self.rat,
                                     partitioning_technique=HoldOutPartitioning(random_state=42),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkSVC()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkKNN(2))],
                                     items_directory=items_dir,
                                     output_folder="my_experiment")

        cbe.predict(num_cpus=num_cpus)

        self.assertTrue(os.path.isdir("my_experiment"))

        # since no algorithm is a score prediction one, only train and test split will be present
        # and no other thing
        contents_directory = os.listdir("my_experiment")

        contents_directory.remove("HoldOutPartitioning_train_split0.csv")
        contents_directory.remove("HoldOutPartitioning_test_split0.csv")

        self.assertTrue(len(contents_directory) == 0)

    def test_predict_raise_error(self):
        cbe = ContentBasedExperiment(self.rat,
                                     partitioning_technique=HoldOutPartitioning(random_state=42),
                                     algorithm_list=[CentroidVector({'Plot': 0}, similarity=CosineSimilarity()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkSVC()),
                                                     ClassifierRecommender({'Plot': 0}, classifier=SkKNN(2))],
                                     items_directory=items_dir,
                                     output_folder="my_experiment")

        with self.assertRaises(NotPredictionAlg):
            cbe.predict(num_cpus=num_cpus, skip_alg_error=False)

    def doCleanups(self) -> None:
        # clean folders created
        shutil.rmtree("my_experiment", ignore_errors=True)


class TestGraphBasedExperiment(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        rat = Ratings(CSVFile(rat_path))
        cls.rat = rat

        cls.gbe = GraphBasedExperiment(rat,
                                       partitioning_technique=KFoldPartitioning(random_state=42, n_splits=2),
                                       algorithm_list=[NXPageRank(alpha=0.8),
                                                       NXPageRank(alpha=0.9)],
                                       items_directory=items_dir,
                                       item_exo_properties={'dbpedia'},
                                       user_exo_properties={'local'},
                                       output_folder="my_experiment")

    def test_rank(self):
        self.gbe.rank(num_cpus=num_cpus)

        # check dirs have been created
        self.assertTrue(os.path.isdir("my_experiment"))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "NXPageRank_1")))
        self.assertTrue(os.path.isdir(os.path.join("my_experiment", "NXPageRank_2")))

        # check train test splits have been created
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_test_split1.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "KFoldPartitioning_train_split1.csv")))

        # check rank created for nx page rank 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_1", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_1", "rs_rank_split1.csv")))

        # check rank created for nx page rank 2
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_2", "rs_rank_split0.csv")))
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_2", "rs_rank_split1.csv")))

    def test_rank_with_report(self):
        gbe = GraphBasedExperiment(self.rat,
                                   partitioning_technique=KFoldPartitioning(random_state=42, n_splits=2),
                                   algorithm_list=[NXPageRank(alpha=0.8),
                                                   NXPageRank(alpha=0.9)],
                                   items_directory=items_dir,
                                   item_exo_properties={'dbpedia'},
                                   user_exo_properties={'local'},
                                   output_folder="my_experiment",
                                   report=True)

        gbe.rank(num_cpus=num_cpus)

        # check report created for nx page rank 1
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_1", "rs_report.yml")))

        # check report created for nx page rank 2
        self.assertTrue(os.path.isfile(os.path.join("my_experiment", "NXPageRank_2", "rs_report.yml")))

    def test_no_one_predictor(self):

        self.gbe.predict(num_cpus=num_cpus)

        self.assertTrue(os.path.isdir("my_experiment"))

        # since no algorithm is a score prediction one, only train and test split will be present
        # and no other thing
        contents_directory = os.listdir("my_experiment")

        contents_directory.remove("KFoldPartitioning_train_split0.csv")
        contents_directory.remove("KFoldPartitioning_test_split0.csv")
        contents_directory.remove("KFoldPartitioning_train_split1.csv")
        contents_directory.remove("KFoldPartitioning_test_split1.csv")

        self.assertTrue(len(contents_directory) == 0)

    def test_predict_raise_error(self):
        with self.assertRaises(NotPredictionAlg):
            self.gbe.predict(num_cpus=num_cpus, skip_alg_error=False)

    def doCleanups(self) -> None:
        # clean folders created
        shutil.rmtree("my_experiment", ignore_errors=True)


if __name__ == '__main__':
    unittest.main()



from typing import Union, List
from nltk import data, download, sent_tokenize, word_tokenize

# nltk corpus
corpus_downloaded = False


def check_tokenized(text):
    """
    Tokenizes a text
    """
    if type(text) is str:
        global corpus_downloaded

        if not corpus_downloaded:
            try:
                data.find('punkt')
            except LookupError:
                download('punkt')

            corpus_downloaded = True

        text = word_tokenize(text)

    return text


def check_not_tokenized(text):
    """
    Untokenizes a tokenized text
    """
    if type(text) is list:
        text = ' '.join(text)

    return text


def tokenize_in_sentences(text: Union[List[str], str]):
    """
    Tokenizes a text into sentences
    """
    global corpus_downloaded

    if not corpus_downloaded:
        try:
            data.find('punkt')
        except LookupError:
            download('punkt')

        corpus_downloaded = True

    return sent_tokenize(check_not_tokenized(text))



from unittest import TestCase
import os

from clayrs.content_analyzer.content_representation.content import FeaturesBagField
from clayrs.content_analyzer.raw_information_source import JSONFile
from clayrs.content_analyzer.field_content_production_techniques import PyWSDSynsetDocumentFrequency
from test import dir_test_files

file_path = os.path.join(dir_test_files, "movies_info_reduced.json")


class TestSynsetDocumentFrequency(TestCase):
    def test_produce_content(self):
        technique = PyWSDSynsetDocumentFrequency()

        features_bag_list = technique.produce_content("Plot", [], JSONFile(file_path))

        self.assertEqual(len(features_bag_list), 20)
        self.assertIsInstance(features_bag_list[0], FeaturesBagField)



from __future__ import annotations
import itertools
from typing import Set, Dict, List, TYPE_CHECKING
from collections import Counter

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings


def get_item_popularity(original_ratings: Ratings) -> Dict:
    """
    Compute item popularity for each item_id in the `original_ratings` parameter

    The popularity of an item is defined as the number of times it is rated in the `original_ratings` parameter
    divided by the total number of users in the `original_ratings`.

    Args:
        original_ratings: `Ratings` object used to compute popularity for each item

    Returns:
        Python dictionary containing popularity computed for each item in the `original_ratings` parameter
    """
    n_users = len(set(original_ratings.user_id_column))

    pop_by_item = {item_id: count / n_users for item_id, count in Counter(original_ratings.item_id_column).items()}

    return pop_by_item


def get_most_popular_items(pop_by_item: Dict, top_n_percentage: float = 0.2) -> Set[str]:
    """
    Find the set of *most popular items*, which are the first `top_n_percentage`% items of all items ordered in a
    descending order by popularity.

    Args:
        pop_by_item: Python dictionary containing popularity computed for each item in the `original_ratings` parameter
        top_n_percentage: How many (in percentage) *most popular items* must be considered. Default is 0.2

    Returns:
        Set of most popular item_ids
    """
    num_of_items = len(pop_by_item)
    top_n_index = round(num_of_items * top_n_percentage)

    sorted_popularities = dict(sorted(pop_by_item.items(), key=lambda item: item[1], reverse=True))
    most_common = dict(itertools.islice(sorted_popularities.items(), top_n_index))

    # only item ids
    most_popular_items = set(most_common.keys())

    return most_popular_items


def pop_ratio_by_user(score_frame: Ratings, most_pop_items: Set[str]) -> Dict:
    r"""
    Compute popularity ratio for each user. $Popularity\_ratio$ for a single user $u$ is defined as:

    $$
    Popularity\_ratio_u = n\_most\_popular\_items\_rated_u / n\_items\_rated_u
    $$

    Args:
        score_frame: `Ratings` object used to compute popularity ratio for each user
        most_pop_items: set of most popular 'item_id' labels

    Returns:
        Python dictionary containing as keys each user id and as value the popularity ratio of each user
    """
    # Splitting users by popularity
    users = set(score_frame.user_id_column)

    popularity_ratio_by_user = {}

    for user in users:
        # filters by the current user and returns all the items he has rated
        user_ratings = score_frame.get_user_interactions(user)
        rated_items = set([user_interaction.item_id for user_interaction in user_ratings])
        # intersects rated_items with popular_items
        popular_rated_items = rated_items.intersection(most_pop_items)
        popularity_ratio = len(popular_rated_items) / len(rated_items)

        popularity_ratio_by_user[user] = popularity_ratio

    return popularity_ratio_by_user


def get_avg_pop(items: List, pop_by_item: Dict) -> float:
    """
    Get the average popularity of the given items list

    Args:
        items: A list containing *item ids*
        pop_by_item: Python dictionary containing popularity computed for each item in the `original_ratings` parameter

    Returns:
        The average popularity of the input `items` list
    """

    popularities = [pop_by_item.get(item, 0.0) for item in items]

    return sum(popularities) / len(items)



from unittest import TestCase
import os
import pathlib as pl

from clayrs.content_analyzer.embeddings.embedding_learner import GensimFastText
from clayrs.content_analyzer.information_processor.nltk import NLTK
from clayrs.content_analyzer.raw_information_source import JSONFile
from test import dir_test_files

file_path = os.path.join(dir_test_files, 'movies_info_reduced.json')


class TestGensimFastText(TestCase):
    def test_fit(self):
        model_path = "./model_test_FastText"
        learner = GensimFastText(model_path, True)
        learner.fit(source=JSONFile(file_path), field_list=["Plot", "Genre"], preprocessor_list=[NLTK()])
        model_path += ".kv"

        self.assertEqual(learner.get_embedding("ace").any(), True)
        self.assertEqual(pl.Path(model_path).resolve().is_file(), True)





from abc import ABC, abstractmethod
from typing import List


class InformationProcessor(ABC):
    """
    Abstract class that generalizes data processing.
    """

    @abstractmethod
    def process(self, field_data):
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class ImageProcessor(InformationProcessor):
    """
    Abstract class for image processing.
    """
    @abstractmethod
    def process(self, field_data):
        raise NotImplementedError


class AudioProcessor(InformationProcessor):
    """
    Abstract class for audio processing.
    """
    @abstractmethod
    def process(self, field_data):
        raise NotImplementedError


class TextProcessor(InformationProcessor):
    """
    Abstract class for raw text processing.
    """

    @staticmethod
    def list_to_string(text: List[str]) -> str:
        """
        Convert list of str in str
        Args: text (str): list of str
        Returns: str sentence
        """
        string_text = ' '.join([str(elem) for elem in text])
        return string_text

    @staticmethod
    def string_to_list(text) -> List[str]:
        """
        Covert str in list of str
        Args:
            text (str): str sentence

        Returns List <str>: List of words
        """
        list_text = list(text.split(" "))
        return list_text

    @abstractmethod
    def process(self, field_data):
        raise NotImplementedError


class NLP(TextProcessor):
    """
    Class for processing a text via Natural Language Processing.

    """

    @abstractmethod
    def process(self, field_data) -> List[str]:
        """
        Apply on the original text the required preprocessing steps
        Args:
            field_data: text on which NLP with specified phases will be applied

        Returns:
            list<str>: The text, after being processed with the specified NLP pipeline,
                is splitted in single words that are put into a list. The splitting is executed
                even if none of the preprocessing steps is computed.
        """
        raise NotImplementedError



from clayrs.recsys.graphs.graph import PropertyNode, UserNode, ItemNode
from clayrs.recsys.graphs.nx_implementation.nx_tripartite_graphs import NXTripartiteGraph
import os

from clayrs.utils import load_content_instance
from test import dir_test_files
from test.recsys.graphs.test_networkx_implementation.test_nx_bipartite_graphs import TestNXBipartiteGraph, rat, \
    rat_timestamp

ratings_filename = os.path.join(dir_test_files, 'new_ratings_small.csv')
movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')
users_dir = os.path.join(dir_test_files, 'complex_contents', 'users_codified/')


class TestNXTripartiteGraph(TestNXBipartiteGraph):

    def setUp(self) -> None:
        # graphs that will be used for testing
        self.g: NXTripartiteGraph = NXTripartiteGraph(rat,
                                                      item_contents_dir=movies_dir,
                                                      item_exo_properties={'dbpedia': ['film director',
                                                                                       'runtime (m)']})

        self.graph_custom_label: NXTripartiteGraph = NXTripartiteGraph(rat, link_label='my_label',
                                                                       item_contents_dir=movies_dir,
                                                                       item_exo_properties={'dbpedia': ['film director',
                                                                                                        'runtime (m)']})

        self.graph_timestamp: NXTripartiteGraph = NXTripartiteGraph(rat_timestamp,
                                                                    item_contents_dir=movies_dir,
                                                                    item_exo_properties={'dbpedia': ['film director',
                                                                                                     'runtime (m)']})

        # this will be empty even if other attributes are specified since ratings are missing
        self.empty_graph: NXTripartiteGraph = NXTripartiteGraph(item_contents_dir=movies_dir,
                                                                item_exo_properties={'dbpedia'})

        # item_exo_properties set but no item_contents_dir specified
        self.graph_missing_item_dir_parameter = NXTripartiteGraph(rat, item_exo_properties={'dbpedia': ['film director',
                                                                                                        'runtime (m)']})

        # item_contents_dir set but no item_exo_properties_specified specified
        self.graph_missing_item_prop_parameter = NXTripartiteGraph(rat, item_contents_dir=movies_dir)

    def test_graph_creation(self):
        # the super class test will check if every user and item have a link
        # as they are present in the ratings frame
        super().test_graph_creation()

        # here we test if item nodes are linked to their exogenous property as specified in the constructor
        for item_node in self.g.item_nodes:
            loaded_item = load_content_instance(movies_dir, item_node.value)
            exogenous_representation: dict = loaded_item.get_exogenous_representation("dbpedia").value
            director_prop_expected = exogenous_representation.get("film director", [])
            runtime_prop_expected = exogenous_representation.get("runtime (m)", [])

            if not isinstance(director_prop_expected, list):
                director_prop_expected = [director_prop_expected]

            if not isinstance(runtime_prop_expected, list):
                runtime_prop_expected = [runtime_prop_expected]

            for director in director_prop_expected:
                self.assertTrue(PropertyNode(director) in self.g.property_nodes)
                result_link_data = self.g.get_link_data(item_node, PropertyNode(director))
                expected_link_data = {'label': 'film director'}

                self.assertEqual(expected_link_data, result_link_data)

            for runtime in runtime_prop_expected:
                self.assertTrue(PropertyNode(runtime) in self.g.property_nodes)
                result_link_data = self.g.get_link_data(item_node, PropertyNode(runtime))
                expected_link_data = {'label': 'runtime (m)'}

                self.assertEqual(expected_link_data, result_link_data)

    def test_graph_creation_missing_parameter(self):

        self.assertTrue(len(self.graph_missing_item_dir_parameter.user_nodes) != 0)
        self.assertTrue(len(self.graph_missing_item_dir_parameter.item_nodes) != 0)
        # warning is printed and no prop will be loaded
        self.assertTrue(len(self.graph_missing_item_dir_parameter.property_nodes) == 0)

        self.assertTrue(len(self.graph_missing_item_dir_parameter.user_nodes) != 0)
        self.assertTrue(len(self.graph_missing_item_dir_parameter.item_nodes) != 0)
        # warning is printed and no prop will be loaded
        self.assertTrue(len(self.graph_missing_item_dir_parameter.property_nodes) == 0)

    def test_graph_creation_empty(self):
        super().test_graph_creation_empty()

        self.assertTrue(len(self.empty_graph.property_nodes) == 0)

    def test_add_property(self):
        # Add 'property' node
        self.assertFalse(PropertyNode('Nolan') in self.g.property_nodes)
        self.g.add_node(PropertyNode('Nolan'))
        self.assertTrue(self.g.node_exists(PropertyNode('Nolan')))
        self.assertTrue(PropertyNode('Nolan') in self.g.property_nodes)

        # Add a list of 'property' nodes
        list_nodes = [PropertyNode('prop1'), PropertyNode('prop2'), PropertyNode('prop3')]
        self.g.add_node(list_nodes)
        for n in list_nodes:
            self.assertTrue(self.g.node_exists(n))
            self.assertTrue(n in self.g.property_nodes)

        # Add 'property' node but it already exists as
        # a 'user' node
        self.g.add_node(UserNode('0'))
        self.assertTrue(self.g.node_exists(UserNode('0')))
        self.assertTrue(UserNode('0') in self.g.user_nodes)
        self.g.add_node(PropertyNode('0'))
        self.assertTrue(self.g.node_exists(PropertyNode('0')))
        self.assertTrue(self.g.node_exists(UserNode('0')))
        self.assertTrue(UserNode('0') in self.g.user_nodes)
        self.assertTrue(PropertyNode('0') in self.g.property_nodes)

    def test_add_link_item_prop_existent(self):
        # Link existent 'item' node to an existent 'prop' node
        item_node = ItemNode('Tenet')
        prop_node = PropertyNode('Nolan')
        self.g.add_node(item_node)
        self.g.add_node(prop_node)
        self.assertIsNone(self.g.get_link_data(item_node, prop_node))
        self.g.add_link(item_node, prop_node, timestamp='now', label='directed_by')
        expected = {'timestamp': 'now', 'label': 'directed_by'}
        result = self.g.get_link_data(item_node, prop_node)
        self.assertEqual(expected, result)

        # Link list of existent 'item' nodes to a list of existent 'property' nodes
        items_list = [ItemNode('i1_list'), ItemNode('i2_list'), ItemNode('i3_list')]
        self.g.add_node(items_list)
        props_list = [PropertyNode('p1_list'), PropertyNode('p2_list'), PropertyNode('p3_list')]
        self.g.add_node(props_list)
        self.g.add_link(items_list, props_list, 0.5)
        for user, item in zip(items_list, props_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

        # Link existent 'property' node to an existent 'item' node
        item_node = ItemNode('Tenet')
        prop_node = PropertyNode('Nolan')
        self.g.add_node(item_node)
        self.g.add_node(prop_node)
        self.assertIsNone(self.g.get_link_data(prop_node, item_node))
        self.g.add_link(prop_node, item_node)
        self.assertIsNotNone(self.g.get_link_data(item_node, prop_node))

        # Link list of existent 'prop' nodes to a list of existent 'item' nodes
        items_list = [ItemNode('i1_list'), ItemNode('i2_list'), ItemNode('i3_list')]
        self.g.add_node(items_list)
        users_list = [UserNode('u1_list'), UserNode('u2_list'), UserNode('u3_list')]
        self.g.add_node(users_list)
        self.g.add_link(items_list, users_list, 0.5)
        for item, user in zip(items_list, users_list):
            result = self.g.get_link_data(item, user)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    def test_add_link_item_prop_non_existent(self):
        # Link non-existent 'user' node and non-existent 'item' node,
        # so both nodes are created and then linked
        item_new = ItemNode('i_new')
        prop_new = PropertyNode('p_new')
        self.assertFalse(self.g.node_exists(item_new))
        self.assertFalse(self.g.node_exists(prop_new))
        self.g.add_link(item_new, prop_new, 0.5)
        self.assertTrue(self.g.node_exists(item_new))
        self.assertTrue(self.g.node_exists(prop_new))
        self.assertIsNotNone(self.g.get_link_data(item_new, prop_new))

        # Link non-existent 'user' node list and non-existent 'item' node list,
        # so all nodes of the two lists are created and then linked
        item_new_list = [ItemNode('i_new_new1'), ItemNode('i_new_new2')]
        prop_new_list = [PropertyNode('u_new_new1'), PropertyNode('u_new_new2')]
        for item in item_new_list:
            self.assertFalse(self.g.node_exists(item))
        for prop in prop_new_list:
            self.assertFalse(self.g.node_exists(prop))

        self.g.add_link(item_new_list, prop_new_list, 0.5)

        for user, item in zip(item_new_list, prop_new_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    def test_add_link_raise_error(self):
        # test link existent 'user' node to existent 'property' node
        self.g.add_node(UserNode('u1'))
        self.g.add_node(PropertyNode('Nolan'))
        with self.assertRaises(ValueError):
            self.g.add_link(UserNode('u1'), PropertyNode('Nolan'), weight=0.5, label='Friend')

        # test link 'property' node to 'user' node
        self.g.add_node(UserNode('u1'))
        self.g.add_node(PropertyNode('Nolan'))
        with self.assertRaises(ValueError):
            self.g.add_link(PropertyNode('Nolan'), UserNode('u1'), weight=0.5, label='Friend')

        # test link list of non-existent user node to list of non-existent property node
        user_list = [UserNode('u_new'), UserNode('u_new1')]
        prop_list = [PropertyNode('p_new'), PropertyNode('p_new1')]
        for n in user_list:
            self.assertFalse(self.g.node_exists(n))
        for n in prop_list:
            self.assertFalse(self.g.node_exists(n))

        with self.assertRaises(ValueError):
            self.g.add_link(user_list, prop_list, weight=0.5, label="PropertyNew")

    def test_add_node_with_prop(self):
        # add 'item' node and only properties 'film_director' to the graph
        self.assertFalse(ItemNode('tt0114709') in self.g.item_nodes)
        self.assertFalse(PropertyNode('http://dbpedia.org/resource/John_Lasseter') in self.g.property_nodes)
        self.g.add_node_with_prop(ItemNode('tt0114709'), {'dbpedia': ['film director']}, movies_dir)
        self.assertTrue(ItemNode('tt0114709') in self.g.item_nodes)
        self.assertTrue(PropertyNode('http://dbpedia.org/resource/John_Lasseter') in self.g.property_nodes)
        self.assertIsNotNone(self.g.get_link_data(ItemNode('tt0114709'),
                                                  PropertyNode('http://dbpedia.org/resource/John_Lasseter')))

        self.g.remove_node(ItemNode('tt0114709'))
        self.g.remove_node(PropertyNode('http://dbpedia.org/resource/John_Lasseter'))

        # add 'item' node and only properties 'film_director' to the graph with different id from the file name
        self.assertFalse(ItemNode('different') in self.g.item_nodes)
        self.assertFalse(PropertyNode('http://dbpedia.org/resource/John_Lasseter') in self.g.property_nodes)
        self.g.add_node_with_prop(ItemNode('different'), {'dbpedia': ['film director']}, movies_dir,
                                  'tt0114709')
        self.assertTrue(ItemNode('different') in self.g.item_nodes)
        self.assertTrue(PropertyNode('http://dbpedia.org/resource/John_Lasseter') in self.g.property_nodes)
        self.assertIsNotNone(self.g.get_link_data(ItemNode('different'),
                                                  PropertyNode('http://dbpedia.org/resource/John_Lasseter')))

        self.g.remove_node(ItemNode('different'))
        self.g.remove_node(PropertyNode('http://dbpedia.org/resource/John_Lasseter'))

        # add 'item' node and all its properties
        loaded_item = load_content_instance(movies_dir, 'tt0114709')
        exogenous_representation: dict = loaded_item.get_exogenous_representation("dbpedia").value

        self.assertFalse(ItemNode('tt0114709') in self.g.item_nodes)
        self.assertFalse(PropertyNode('http://dbpedia.org/resource/John_Lasseter') in self.g.property_nodes)
        self.g.add_node_with_prop(ItemNode('tt0114709'), {'dbpedia'}, movies_dir)
        self.assertTrue(ItemNode('tt0114709') in self.g.item_nodes)

        for prop_label in exogenous_representation.keys():
            prop_val_expected = exogenous_representation.get(prop_label, [])

            if not isinstance(prop_val_expected, list):
                prop_val_expected = [prop_val_expected]

            for prop_val in prop_val_expected:
                self.assertTrue(PropertyNode(prop_val) in self.g.property_nodes)
                result_link_data_list = self.g.get_link_data(ItemNode('tt0114709'), PropertyNode(prop_val))

                # we make sure that this is a property of the exogenous representation
                self.assertIn(result_link_data_list['label'], exogenous_representation)

    def test_add_raise_error(self):
        # Try to add 'user' node with its prop
        with self.assertRaises(ValueError):
            self.g.add_node_with_prop(UserNode('user'), {'representation_id'}, users_dir)

    # def test_convert_to_dataframe(self):
    #     converted_df = self.g.convert_to_dataframe()
    #     self.assertNotIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertIsInstance(user, Node)
    #         self.assertIsInstance(item, Node)
    #
    #     converted_df = converted_df.query('to_id not in @self.g.property_nodes')
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))
    #
    #     converted_df = self.g.convert_to_dataframe(only_values=True, with_label=True)
    #     self.assertIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertNotIsInstance(user, Node)
    #         self.assertNotIsInstance(item, Node)
    #
    #     converted_df = converted_df.query('to_id not in @self.g.property_nodes')[['from_id', 'to_id', 'score']]
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))



from __future__ import annotations
from abc import abstractmethod
from typing import TYPE_CHECKING

import numpy as np
import torch
import transformers
from transformers import AutoModel, AutoTokenizer

if TYPE_CHECKING:
    from clayrs.content_analyzer.embeddings.embedding_loader.vector_strategy import VectorStrategy
    from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique \
        import CombiningTechnique

from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import SentenceEmbeddingLoader
from clayrs.content_analyzer.embeddings.embedding_loader.vector_strategy import CatStrategy
from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import Centroid


class Transformers(SentenceEmbeddingLoader):
    """
    Abstract class for Transformers
    """

    def __init__(self, model_name: str = 'bert-base-uncased',
                 vec_strategy: VectorStrategy = CatStrategy(1),
                 pooling_strategy: CombiningTechnique = Centroid()):
        self._model = None
        self._tokenizer = AutoTokenizer.from_pretrained(model_name)
        self._name_model = model_name
        self._vec_strategy = vec_strategy
        self._last_interesting_layers = vec_strategy.last_interesting_layers
        self._pooling_strategy = pooling_strategy
        super().__init__(model_name)

    def load_model(self):
        # we disable logger info on the load of the _model
        original_verb = transformers.logging.get_verbosity()
        transformers.logging.set_verbosity_error()
        try:
            self._model = AutoModel.from_pretrained(self._name_model, output_hidden_states=True)

            transformers.logging.set_verbosity(original_verb)
            return self._model
        except (OSError, AttributeError):
            raise FileNotFoundError

    def get_vector_size(self) -> int:
        if isinstance(self._vec_strategy, CatStrategy):
            return self.model.config.hidden_size * self._last_interesting_layers
        else:
            return self.model.config.hidden_size

    def get_embedding(self, sentence: str) -> np.ndarray:
        token_vecs = self.get_embedding_token(sentence)
        sentence_vec = self._pooling_strategy.combine(token_vecs)
        return sentence_vec

    @abstractmethod
    def get_embedding_token(self, sentence: str) -> np.ndarray:
        raise NotImplementedError

    def get_sentence_token(self, sentence: str):
        marked_text = "[CLS] " + sentence + " [SEP]"
        tokenized_text = self.model.tokenize(marked_text)
        return tokenized_text

    @abstractmethod
    def __str__(self):
        raise NotImplementedError


class BertTransformers(Transformers):
    """
    Class that produces sentences/token embeddings using any Bert model from hugging face.

    Args:
        model_name: Name of the embeddings model to download or path where the model is stored locally
        vec_strategy: Strategy which will be used to combine each output layer to obtain a single one
        pooling_strategy: Strategy which will be used to combine the embedding representation of each token into a
            single one, representing the embedding of the whole sentence
    """
    def __init__(self, model_name: str = 'bert-base-uncased',
                 vec_strategy: VectorStrategy = CatStrategy(1),
                 pooling_strategy: CombiningTechnique = Centroid()):
        super().__init__(model_name, vec_strategy, pooling_strategy)

    def get_embedding_token(self, sentence: str) -> np.ndarray:
        encoded = self._tokenizer(sentence, truncation=True, return_tensors='pt')

        with torch.no_grad():
            model_output = self.model(**encoded)
            hidden_states = model_output['hidden_states']

        token_embeddings = torch.stack(hidden_states, dim=0)
        token_embeddings = torch.squeeze(token_embeddings, dim=1)
        token_embeddings = token_embeddings.permute(1, 0, 2)

        token_vecs = self._vec_strategy.build_embedding(token_embeddings)

        return token_vecs

    def __str__(self):
        return "BertTransformers"

    def __repr__(self):
        return f"BertTransformers(model_name={self._name_model}, " \
               f"vec_strategy={self._vec_strategy}, " \
               f"pooling_strategy={self._pooling_strategy})"


class T5Transformers(Transformers):
    """
    Class that produces sentences/token embeddings using sbert.

    Args:
        model_name: Name of the embeddings model to download or path where the model is stored locally
        vec_strategy: Strategy which will be used to combine each output layer to obtain a single one
        pooling_strategy: Strategy which will be used to combine the embedding representation of each token into a
            single one, representing the embedding of the whole sentence
    """
    def __init__(self, model_name: str = 't5-small',
                 vec_strategy: VectorStrategy = CatStrategy(1),
                 pooling_strategy: CombiningTechnique = Centroid()):
        super().__init__(model_name, vec_strategy, pooling_strategy)

    def get_embedding_token(self, sentence: str) -> np.ndarray:
        encoded = self._tokenizer(sentence, truncation=True, return_tensors='pt')

        with torch.no_grad():
            model_output = self.model.encoder(**encoded)
            hidden_states = model_output['hidden_states']

        token_embeddings = torch.stack(hidden_states, dim=0)
        token_embeddings = torch.squeeze(token_embeddings, dim=1)
        token_embeddings = token_embeddings.permute(1, 0, 2)

        token_vecs = self._vec_strategy.build_embedding(token_embeddings)

        return token_vecs

    def __str__(self):
        return "T5Transformers"

    def __repr__(self):
        return f"T5Transformers(model_name={self._name_model}, " \
               f"vec_strategy={self._vec_strategy}, " \
               f"pooling_strategy={self._pooling_strategy})"



from .classification_metrics import Precision, PrecisionAtK, RPrecision, Recall, RecallAtK, \
    FMeasure, FMeasureAtK
from .error_metrics import MAE, MSE, RMSE
from .fairness_metrics import GiniIndex, DeltaGap, PredictionCoverage, CatalogCoverage
from .plot_metrics import PopRatioProfileVsRecs, PopRecsCorrelation, LongTailDistr
from .ranking_metrics import NDCG, NDCGAtK, MRR, MRRAtK, Correlation, MAP



import os

import pandas as pd

from clayrs.content_analyzer import Ratings
from clayrs.recsys import NXFullGraph
from test import dir_test_files
from unittest import TestCase

ratings_filename = os.path.join(dir_test_files, 'new_ratings_small.csv')
movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')
users_dir = os.path.join(dir_test_files, 'complex_contents', 'users_codified/')

rat = pd.DataFrame.from_dict({'from_id': ["1", "1", "2", "2", "2", "3", "4", "4"],
                              'to_id': ["tt0112281", "tt0112302", "tt0112281", "tt0112346",
                                        "tt0112453", "tt0112453", "tt0112346", "tt0112453"],
                              'score': [0.8, 0.7, -0.4, 1.0, 0.4, 0.1, -0.3, 0.7]})
rat = Ratings.from_dataframe(rat)


class TestGraph(TestCase):
    def setUp(self) -> None:
        # we need to instantiate a sublass of Graph since it's an abstract class to test its methods
        self.g: NXFullGraph = NXFullGraph(rat,
                                          item_contents_dir=movies_dir,
                                          item_exo_properties={'dbpedia': ['film director',
                                                                           'runtime (m)']},

                                          # It's the column in the users .DAT which identifies the gender
                                          user_exo_properties={'local': '1'},
                                          user_contents_dir=users_dir)

    def test_to_ratings(self):
        converted_rat = self.g.to_ratings()

        # check that original ratings and converted ratings are equal

        self.assertEqual(set(rat.user_id_column), set(converted_rat.user_id_column))
        self.assertEqual(set(rat.item_id_column), set(converted_rat.item_id_column))
        self.assertEqual(set(rat.score_column), set(converted_rat.score_column))
        self.assertEqual(set(rat.timestamp_column), set(converted_rat.timestamp_column))

        for user in rat.user_id_column:
            user_rat = rat.get_user_interactions(user)
            user_converted_rat = converted_rat.get_user_interactions(user)

            self.assertCountEqual(user_rat, user_converted_rat)



from __future__ import annotations
import abc
import os
import shutil
import sys
from abc import ABC
from typing import List, Union, Dict, Callable, TYPE_CHECKING

# fix circular import, for the future: move Ratings class to the RecSys module
if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings
    from clayrs.evaluation.metrics.metrics import Metric
    from clayrs.recsys.algorithm import Algorithm
    from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm
    from clayrs.recsys.graph_based_algorithm.graph_based_algorithm import GraphBasedAlgorithm
    from clayrs.recsys.partitioning import Partitioning
    from clayrs.recsys.methodology import Methodology

from clayrs.evaluation import EvalModel
from clayrs.recsys import ContentBasedRS, GraphBasedRS, NXFullGraph, UserNode, ItemNode
from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg
from clayrs.utils import Report
from clayrs.utils.const import logger
from clayrs.utils.save_content import get_valid_dirname
from clayrs.recsys.methodology import TestRatingsMethodology


class Experiment(ABC):
    """
    Abstract `Experiment` class

    It provides an easy interface to perform a complete experiment by comparing different algorithms, starting from
    splitting the dataset to evaluating predictions computed.
    It is also capable of producing a `yml` report for both the recsys phase and the evaluation phase.

    Both the evaluation phase and the report are optional and are produced only if specified.

    All the results (split, ranking, evaluation results, etc.) will be saved in the folder specified with the
    `output_folder` parameter. For each algorithm a different sub-folder will be created and named after the algorithm:

    * If multiple instances of the same algorithm are present in the `algorithm_list`, sub-folders will be disambiguated
        depending on the order of execution (`algName_1`, `algName_2`, `algName_3`, etc.)

    !!! info

        Please remember that by default if a folder with same name of the `output_folder` parameter is present,
        the experiment won't run and an exception will be raised. To overcome this, simply set the `overwrite_if_exists`
        parameter to True or change the `output_folder`.

    Args:
        original_ratings: Ratings object containing original interactions between users and items

        partitioning_technique: Partitioning object which specifies how the original ratings should be split

        algorithm_list: List of algorithm for which the experiment will be conducted

        items_directory: Path to the folder containing serialized complexly represented items

        users_directory: Path to the folder containing serialized complexly represented items

        metric_list: List of metric with which predictions computed by the RecSys will be evaluated

        report: If `True`, a `yml` report will be produced for the recsys phase. It will be also produced for the
            evaluation phase, but only if the `metric_list` parameter is set

        output_folder: Path of the folder where all the results of the experiment will be saved

        overwrite_if_exists: If `True` and the path set in `output_folder` points to an already existent directory,
            it will be deleted and replaced by the folder containing the experiment results
    """

    def __init__(self, original_ratings: Ratings,
                 partitioning_technique: Partitioning,
                 algorithm_list: List[Algorithm],
                 items_directory: str,
                 users_directory: str = None,
                 metric_list: List[Metric] = None,
                 report: bool = False,
                 output_folder: str = "experiment_result",
                 overwrite_if_exists: bool = False):

        self.original_ratings = original_ratings
        self.algorithm_list = algorithm_list
        self.pt = partitioning_technique
        self.items_directory = items_directory
        self.users_directory = users_directory
        self.metrics = metric_list
        self.report = report
        self.output_folder = output_folder
        self.overwrite_if_exists = overwrite_if_exists

    def main_experiment(self, alg_function: Callable, skip_alg_error: bool = True) -> None:
        """
        Main function which will perform the whole experiment

        It is responsible for splitting the dataset, computing ranking/score predictions, evaluating them, etc.
        The key parameter is `alg_function`:
        each subclass of Experiment should call this method by passing the appropriate callable (function) which should
        expect:

        * The current split number,
        * The current algorithm,
        * The current train set,
        * The current test set,
        * The current dirname (the directory where all the results of the current algorithm should be saved)

        The `alg_function` should return a Ratings object (or one of its subclasses)

        With this abstraction, all the subclasses should just define the `alg_function` function and just call this
        method

        Args:
            alg_function: Callable which perform the ranking/score prediction. Should return a Ratings object (or one
                of its subclasses)
            skip_alg_error: Boolean which defines the behaviour of the experiment when an algorithm raises an error.
                Useful for example when a pure ranking algorithm is asked to perform score prediction, if is set to
                `True` it will just skip to the next algorithm, otherwise the exception raised will be re-raised
        """

        try:
            os.makedirs(self.output_folder)
        except FileExistsError:
            if self.overwrite_if_exists is True:
                shutil.rmtree(self.output_folder)
                os.makedirs(self.output_folder)
            else:
                raise FileExistsError(f"Folder {self.overwrite_if_exists} already present!\n"
                                      "Delete it and run the experiment again or set `overwrite_if_exists` parameter "
                                      "to True!") from None

        train_set_list, test_set_list = self.pt.split_all(self.original_ratings)

        for i, (train_set, test_set) in enumerate(zip(train_set_list, test_set_list)):
            train_set.to_csv(self.output_folder, file_name=f"{str(self.pt)}_train_split{i}")
            test_set.to_csv(self.output_folder, file_name=f"{str(self.pt)}_test_split{i}")

        for alg in self.algorithm_list:

            # Control variable that lets you skip directly to the next algorithm
            # when the NotPredictionAlg exception is raised but should be skipped
            skip_to_next_alg = False

            print(file=sys.stderr)  # just a simple space
            logger.info(f"******* Processing alg {alg} *******")

            all_alg_res = []

            dirname = get_valid_dirname(self.output_folder,
                                        str(alg),
                                        overwrite=False,
                                        start_from_1=True,
                                        style="underscore")

            os.makedirs(os.path.join(self.output_folder, dirname))

            for i, (train_set, test_set) in enumerate(zip(train_set_list, test_set_list)):

                try:

                    result = alg_function(i, alg, train_set, test_set, dirname)

                    all_alg_res.append(result)

                except NotPredictionAlg as e:
                    os.rmdir(f"{self.output_folder}/{dirname}")

                    if skip_alg_error:
                        logger.warning(str(e) + "\nThe algorithm will be skipped")
                        skip_to_next_alg = True
                        break  # we skip directly to the next algorithm to perform
                    else:
                        raise e

            if not skip_to_next_alg:
                if self.metrics is not None:
                    em = EvalModel(pred_list=all_alg_res,
                                   truth_list=test_set_list,
                                   metric_list=self.metrics)

                    sys_df, users_df = em.fit()

                    sys_df.to_csv(os.path.join(self.output_folder, dirname, 'eva_sys_results.csv'))
                    users_df.to_csv(os.path.join(self.output_folder, dirname, 'eva_users_results.csv'))

                    if self.report:
                        Report(output_dir=f"{self.output_folder}/{dirname}").yaml(eval_model=em)

                print(file=sys.stderr)  # just a simple space
                logger.info(f"Results saved in '{self.output_folder}/{dirname}'")

    @abc.abstractmethod
    def predict(self, methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 0,
                skip_alg_error: bool = True):
        """
        Abstract method which should perform an experiment where score predictions are involved
        """
        raise NotImplementedError

    @abc.abstractmethod
    def rank(self, n_recs: int = 10,
             methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 0):
        """
        Abstract method which should perform an experiment where rankings are involved
        """
        raise NotImplementedError


class ContentBasedExperiment(Experiment):
    """
    The `Experiment` class for ***content based algorithms***

    It provides an easy interface to perform a complete experiment by comparing different cb-algorithms, starting from
    *splitting* the dataset to *evaluating* predictions computed.
    It is also capable of producing a `yml` report for both the recsys phase and the evaluation phase.

    Both the evaluation phase and the report are *optional* and are produced only if specified.

    All the results (*split*, *ranking*, *evaluation results*, etc.) will be saved in the folder specified with the
    `output_folder` parameter. For each algorithm a different sub-folder will be created and named after it:

    * If multiple instances of the same algorithm are present in the `algorithm_list`, sub-folders will be disambiguated
        depending on the order of execution (`algName_1`, `algName_2`, `algName_3`, etc.)

    !!! info

        Please remember that by default if a folder with same name of the `output_folder` parameter is present,
        the experiment won't run and an exception will be raised. To overcome this, simply set the `overwrite_if_exists`
        parameter to `True` or change the `output_folder`.

    Examples:

        Suppose you want to compare:

        * A `CentroidVector` algorithm
        * The `SVC` classifier
        * The `KNN` classifier

        For the three different configuration, an `HoldOut` partitioning technique should be used and results should
        be evaluated on $Precision$ and $Recall$

        ```python
        from clayrs.utils import ContentBasedExperiment
        from clayrs import content_analyzer as ca
        from clayrs import content_analyzer as rs

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        alg1 = rs.CentroidVector({'Plot': 'tfidf'},
                                 similarity=rs.CosineSimilarity()) # (1)

        alg2 = rs.ClassifierRecommender({'Plot': 'tfidf'},
                                        classifier=rs.SkSVC()) # (2)

        alg3 = rs.ClassifierRecommender({'Plot': 'tfidf'},
                                         classifier=rs.SkKNN()) # (3)

        a = ContentBasedExperiment(
            original_ratings=rat,

            partitioning_technique=rs.HoldOutPartitioning(),

            algorithm_list=[alg1, alg2, alg3],

            items_directory=movies_dir,

            metric_list=[eva.Precision(), eva.Recall()]

            output_folder="my_experiment"
        )

        a.rank()
        ```

        1. Results will be saved in *my_experiment/CentroidVector_1*
        2. Results will be saved in *my_experiment/ClassifierRecommender_1*
        3. Results will be saved in *my_experiment/ClassifierRecommender_2*

    Args:
        original_ratings: Ratings object containing original interactions between users and items

        partitioning_technique: Partitioning object which specifies how the original ratings should be split

        algorithm_list: List of **Content Based algorithms** for which the whole experiment will be conducted

        items_directory: Path to the folder containing serialized complexly represented items

        users_directory: Path to the folder containing serialized complexly represented items. Needed only if
            one or more algorithms in `algorithm_list` needs it

        metric_list: List of metric with which predictions computed by the CBRS will be evaluated

        report: If `True`, a `yml` report will be produced for the recsys phase. It will be also produced for the
            evaluation phase, but only if the `metric_list` parameter is set

        output_folder: Path of the folder where all the results of the experiment will be saved

        overwrite_if_exists: If `True` and the path set in `output_folder` points to an already existent directory,
            it will be deleted and replaced by the folder containing the experiment results
    """
    def __init__(self, original_ratings: Ratings,
                 partitioning_technique: Partitioning,
                 algorithm_list: List[ContentBasedAlgorithm],
                 items_directory: str,
                 users_directory: str = None,
                 metric_list: List[Metric] = None,
                 report: bool = False,
                 output_folder: str = "experiment_result",
                 overwrite_if_exists: bool = False):

        super().__init__(original_ratings, partitioning_technique, algorithm_list, items_directory, users_directory,
                         metric_list, report, output_folder, overwrite_if_exists)

    def predict(self, methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 0,
                skip_alg_error: bool = True) -> None:
        """
        Method used to perform an experiment which involves ***score predictions***.

        The method will first split the original ratings passed in the constructor in train and test set, then
        the Recommender System will be fit for each user in the train set.

        If the algorithm can't be fit for some users, a warning message is printed and no score predictions will be
        computed for said user

        !!! info

            **BE CAREFUL**: not all algorithms are able to perform *score prediction*. In case a pure ranking algorithm
            is asked to perform score prediction, the `NotPredictionAlg` will be raised. if the `skip_alg_error` is set
            to `True`, then said exception will be caught, a warning will be printed, and the experiment will go on with
            the next algorithm

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be considered for score
        prediction

        Args:
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.
            skip_alg_error: If set to `True`, a pure ranking algorithm will be skipped and the experiment will continue
                with the following algorithm. Otherwise, the `NotPredictionAlg` exception raised will be re-raised

        Raises:
            NotPredictionAlg: When a pure ranking algorithm is asked to perform score prediction and
                `skip_alg_error` == False
        """
        def cb_fit_predict(split_num, alg, train_set, test_set, dirname):
            cbrs = ContentBasedRS(alg, train_set, self.items_directory)

            predict_alg = cbrs.fit_predict(test_set, methodology=methodology, num_cpus=num_cpus)

            predict_alg.to_csv(f"{self.output_folder}/{dirname}", file_name=f"rs_predict_split{split_num}")

            if self.report:
                Report(output_dir=f"{self.output_folder}/{dirname}").yaml(original_ratings=self.original_ratings,
                                                                          partitioning_technique=self.pt,
                                                                          recsys=cbrs)

            return predict_alg

        self.main_experiment(cb_fit_predict, skip_alg_error=skip_alg_error)

    def rank(self, n_recs: int = 10,
             methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 0):
        """
        Method used to perform an experiment which involves ***rankings***.

        The method will first split the original ratings passed in the constructor in train and test set, then
        the Recommender System will be fit for each user in the train.
        If the algorithm can't be fit for some users, a warning message is printed and no ranking will be
        computed for said user

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be considered as eligible
        for ranking

        Args:
            n_recs: Number of the top items that will be present in the ranking of each user.
                If `None` all candidate items will be returned for the user
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected

        """
        def cb_fit_rank(split_num, alg, train_set, test_set, dirname):
            cbrs = ContentBasedRS(alg, train_set, self.items_directory)

            predict_alg = cbrs.fit_rank(test_set, n_recs=n_recs, methodology=methodology, num_cpus=num_cpus)

            predict_alg.to_csv(f"{self.output_folder}/{dirname}", file_name=f"rs_rank_split{split_num}")

            if self.report:
                Report(output_dir=f"{self.output_folder}/{dirname}").yaml(original_ratings=self.original_ratings,
                                                                          partitioning_technique=self.pt,
                                                                          recsys=cbrs)

            return predict_alg

        self.main_experiment(cb_fit_rank)


class GraphBasedExperiment(Experiment):
    """
    The `Experiment` class for ***graph based algorithms***

    It provides an easy interface to perform a **complete experiment** by comparing different gb-algorithms, starting
    from *splitting* the dataset to *evaluating* predictions computed.

    Every graph based algorithm expects a graph: that's why right before computing ranking/score predictions, a graph
    will be created depending on the current train and test split (if multiple are available):

    * All the nodes from the original graph will be present, The interactions in the test set will be
        missing (It won't be represented as a link between a user node and an item node)

    The class is also capable of producing a `yml` report for both the recsys phase and the evaluation phase.

    Both the evaluation phase and the report are optional and are produced only if specified.

    All the results (*split*, *ranking*, *evaluation results*, etc.) will be saved in the folder specified with the
    `output_folder` parameter. For each algorithm a different sub-folder will be created and named after it:

    * If multiple instances of the same algorithm are present in the `algorithm_list`, sub-folders will be disambiguated
        depending on the order of execution (`algName_1`, `algName_2`, `algName_3`, etc.)

    !!! info

        Please remember that by default if a folder with same name of the `output_folder` parameter is present,
        the experiment won't run and an exception will be raised. To overcome this, simply set the `overwrite_if_exists`
        parameter to `True` or change the `output_folder`.

    Examples:

        Suppose you want to compare:

        * The `PageRank` algorithm with `alpha=0.8`
        * The `PageRank` algorithm with `alpha=0.9`
        * The `Personalized PageRank` algorithm

        For the three different configuration, a `KFold` partitioning technique with three splits should be used and
        results should be evaluated on $Precision$, $Recall$, $NDCG$

        ```python
        from clayrs.utils import GraphBasedExperiment
        from clayrs import content_analyzer as ca
        from clayrs import content_analyzer as rs

        original_rat = ca.Ratings(ca.CSVFile(ratings_path))

        alg1 = rs.NXPageRank(alpha=0.8) # (1)

        alg2 = rs.NXPageRank(alpha=0.9) # (2)

        alg3 = rs.NXPageRank(personalized=True) # (3)

        a = GraphBasedExperiment(
            original_ratings=rat,

            partitioning_technique=rs.KFoldPartitioning(n_splits=3),

            algorithm_list=[alg1, alg2, alg3],

            items_directory=movies_dir,

            metric_list=[eva.Precision(), eva.Recall()]

            output_folder="my_experiment"
        )

        a.rank()
        ```

        1. Results will be saved in *my_experiment/NXPageRank_1*
        2. Results will be saved in *my_experiment/NXPageRank_2*
        3. Results will be saved in *my_experiment/NXPageRank_3*

    Args:
        original_ratings: Ratings object containing original interactions between users and items

        partitioning_technique: Partitioning object which specifies how the original ratings should be split

        algorithm_list: List of **Graph Based algorithms** for which the whole experiment will be conducted

        items_directory: Path to the folder containing serialized complexly represented items with one or more
            exogenous property to load

        item_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
            to load all properties from specific representations, or use a `Dict` if you want to choose which properties
            to load from specific representations

        users_directory: Path to the folder containing serialized complexly represented users with one or more
            exogenous property to load

        user_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
            to load all properties from specific representations, or use a `Dict` if you want to choose which properties
            to load from specific representations

        link_label: If specified, each link between user and item nodes will be labeled with the given label.
            Default is None

        metric_list: List of metric with which predictions computed by the GBRS will be evaluated

        report: If `True`, a `yml` report will be produced for the recsys phase. It will be also produced for the
            evaluation phase, but only if the `metric_list` parameter is set

        output_folder: Path of the folder where all the results of the experiment will be saved

        overwrite_if_exists: If `True` and the path set in `output_folder` points to an already existent directory,
            it will be deleted and replaced by the folder containing the experiment results
    """
    def __init__(self, original_ratings: Ratings,
                 partitioning_technique: Partitioning,
                 algorithm_list: List[GraphBasedAlgorithm],
                 items_directory: str = None,
                 item_exo_properties: Union[Dict, set] = None,
                 users_directory: str = None,
                 user_exo_properties: Union[Dict, set] = None,
                 link_label: str = None,
                 metric_list: List[Metric] = None,
                 report: bool = False,
                 output_folder: str = "experiment_result",
                 overwrite_if_exists: bool = False):

        super().__init__(original_ratings, partitioning_technique, algorithm_list, items_directory, users_directory,
                         metric_list, report, output_folder, overwrite_if_exists)

        self.item_exo_properties = item_exo_properties
        self.user_exo_properties = user_exo_properties
        self.link_label = link_label

    def predict(self, methodology: Union[Methodology, None] = TestRatingsMethodology(),
                num_cpus: int = 0,
                skip_alg_error: bool = True):
        """
        Method used to perform an experiment which involves ***score predictions***.

        The method will first split the original ratings passed in the constructor in train and test set, then
        a graph will be built depending on them:

        * All nodes of the original ratings will be present, but the *links* (***interactions***) that are present in
        the test set will be missing, so to make the training phase *fair*

        !!! info

            **BE CAREFUL**: not all algorithms are able to perform *score prediction*. In case a pure ranking algorithm
            is asked to perform score prediction, the `NotPredictionAlg` will be raised. if the `skip_alg_error` is set
            to `True`, then said exception will be caught, a warning will be printed, and the experiment will go on with
            the next algorithm

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used: so, for each user, items in its test set only will be considered for score
        prediction

        Args:
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.
            skip_alg_error: If set to `True`, a pure ranking algorithm will be skipped and the experiment will continue
                with the following algorithm. Otherwise, the `NotPredictionAlg` exception raised will be re-raised

        Raises:
            NotPredictionAlg: When a pure ranking algorithm is asked to perform score prediction
                and `skip_alg_error` == False
        """
        def gb_fit_predict(split_num, alg, train_set, test_set, dirname):

            graph = NXFullGraph(self.original_ratings,
                                item_contents_dir=self.items_directory,
                                item_exo_properties=self.item_exo_properties,
                                user_contents_dir=self.users_directory,
                                user_exo_properties=self.user_exo_properties,
                                link_label=self.link_label)

            for user, item in zip(test_set.user_id_column, test_set.item_id_column):
                graph.remove_link(UserNode(user), ItemNode(item))

            gbrs = GraphBasedRS(alg, graph)

            predict_alg = gbrs.predict(test_set, methodology=methodology, num_cpus=num_cpus)

            predict_alg.to_csv(f"{self.output_folder}/{dirname}", file_name=f"rs_predict_split{split_num}")

            if self.report:
                Report(output_dir=f"{self.output_folder}/{dirname}").yaml(original_ratings=self.original_ratings,
                                                                          partitioning_technique=self.pt,
                                                                          recsys=gbrs)

            return predict_alg

        self.main_experiment(gb_fit_predict, skip_alg_error=skip_alg_error)

    def rank(self, n_recs: int = 10,
             methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 0):
        """
        Method used to perform an experiment which involves ***rankings***.

        The method will first split the original ratings passed in the constructor in train and test set, then
        a graph will be built depending on them:

        * All nodes of the original ratings will be present, but the *links* (***interactions***) that are present in
        the test set will be missing, so to make the training phase *fair*

        Via the `methodology` parameter you can perform different candidate item selection. By default, the
        `TestRatingsMethodology()` is used, so for each user, items in its test set only will be eligible for ranking

        Args:
            n_recs: Number of the top items that will be present in the ranking of each user.
                If `None` all candidate items will be returned for the user
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        """
        def gb_fit_rank(split_num, alg, train_set, test_set, dirname):

            graph = NXFullGraph(self.original_ratings,
                                item_contents_dir=self.items_directory,
                                item_exo_properties=self.item_exo_properties,
                                user_contents_dir=self.users_directory,
                                user_exo_properties=self.user_exo_properties,
                                link_label=self.link_label)

            for user, item in zip(test_set.user_id_column, test_set.item_id_column):
                graph.remove_link(UserNode(user), ItemNode(item))

            gbrs = GraphBasedRS(alg, graph)

            predict_alg = gbrs.rank(test_set, n_recs=n_recs, methodology=methodology, num_cpus=num_cpus)

            predict_alg.to_csv(f"{self.output_folder}/{dirname}", file_name=f"rs_rank_split{split_num}")

            if self.report:
                Report(output_dir=f"{self.output_folder}/{dirname}").yaml(original_ratings=self.original_ratings,
                                                                          partitioning_technique=self.pt,
                                                                          recsys=gbrs)

            return predict_alg

        self.main_experiment(gb_fit_rank)



from abc import ABC, abstractmethod

import numpy as np
import torch


class VectorStrategy(ABC):
    """
    Useful class in order to abstract the strategies that use more layers in order to obtain an only final
    representation of the model
    """
    def __init__(self, last_interesting_layers: int):
        self.last_interesting_layers = last_interesting_layers

    @abstractmethod
    def build_embedding(self, token_embeddings) -> np.ndarray:
        raise NotImplementedError


class SumStrategy(VectorStrategy):
    """
    Class which sums the `last_interesting_layers` of the output obtained by the Transformer model

    Args:
        last_interesting_layers: Which layer to sum in order to summarize information
    """
    def __init__(self, last_interesting_layers: int):
        super().__init__(last_interesting_layers)

    def build_embedding(self, token_embeddings: torch.Tensor) -> np.ndarray:
        token_vecs_sum = []
        for token in token_embeddings:
            sum_vec = torch.sum(token[-self.last_interesting_layers:], dim=0)
            token_vecs_sum.append(sum_vec)
        return torch.stack(token_vecs_sum).numpy()

    def __str__(self):
        return "SumStrategy"

    def __repr__(self):
        return f"SumStrategy(last_interesting_layers={self.last_interesting_layers})"


class CatStrategy(VectorStrategy):
    """
    Class which concatenate the `last_interesting_layers` of the output obtained by the Transformer model

    Args:
        last_interesting_layers: Which layer to concatenate in order to summarize information
    """
    def __init__(self, last_interesting_layers: int):
        super().__init__(last_interesting_layers)

    def build_embedding(self, token_embeddings: torch.Tensor) -> np.ndarray:
        token_vecs_cat = []
        for token in token_embeddings:
            cat_vec = token[-1]
            for i in range(-2, -self.last_interesting_layers - 1, -1):
                cat_vec = torch.cat((cat_vec, token[i]), dim=0)
            token_vecs_cat.append(cat_vec)
        return torch.stack(token_vecs_cat).numpy()

    def __str__(self):
        return "CatStrategy"

    def __repr__(self):
        return f"CatStrategy(last_interesting_layers={self.last_interesting_layers})"



from textblob import TextBlob

from clayrs.content_analyzer.ratings_manager.score_processor import SentimentAnalysis


class TextBlobSentimentAnalysis(SentimentAnalysis):
    """
    Class that compute sentiment polarity on a textual field using TextBlob library.

    The given score will be in the $[-1.0, 1.0]$ range
    """
    def __init__(self, decimal_rounding: int = None):
        super().__init__(decimal_rounding)

    def __str__(self):
        return "TextBlobSentimentalAnalysis"

    def __repr__(self):
        return f'TextBlobSentimentAnalysis'

    def fit(self, score_data: str) -> float:
        """
        This method calculates the sentiment polarity score on textual reviews

        Args:
            score_data: text for which sentiment polarity must be computed and considered as score

        Returns:
            The sentiment polarity of the textual data in range $[-1.0, 1.0]$
        """
        polarity_score = TextBlob(score_data).sentiment.polarity

        if self.decimal_rounding:
            polarity_score = round(polarity_score, self.decimal_rounding)

        return polarity_score



from unittest import TestCase

from nltk import Tree

from clayrs.content_analyzer.information_processor.nltk import NLTK


class TestNLTK(TestCase):

    def test_process_stopwords(self):
        nltka = NLTK(stopwords_removal=True)
        expected = ["striped", "bats", "hanging", "feet", "best"]
        result = nltka.process("The striped bats are hanging on their feet for the best")
        self.assertEqual(expected, result)

    def test_process_stemming(self):
        # Test for only stemming
        nltka = NLTK(stemming=True)
        expected = ['these', 'unbeliev', 'abnorm', 'object']
        result = nltka.process("These unbelievable abnormous objects")

        self.assertEqual(expected, result)

    def test_process_lemmatization(self):
        nltka = NLTK(lemmatization=True)
        expected = ["The", "strip", "bat", "be", "hang", "on", "their", "foot", "for", "best"]
        result = nltka.process("The striped bats are hanging on their feet for best")
        self.assertEqual(expected, result)

    def test_remove_punctuation(self):
        nltka = NLTK(remove_punctuation=True)
        expected = ["Hello", "there", "How", "are", "you", "I", "m", "fine", "thanks"]
        result = nltka.process("Hello there. How are you? I'm fine, thanks.")
        self.assertEqual(expected, result)

    def test_url_tagging(self):
        nltka = NLTK(url_tagging=True)
        expected = ["The", "striped", "<URL>", "bats", "<URL>", "are", "<URL>", "hanging", "on", "their", "feet", "for",
                    "best", "<URL>"]
        result = nltka.process("The striped http://facebook.com bats https://github.com are "
                               "http://facebook.com hanging on their feet for best http://twitter.it")
        self.assertEqual(expected, result)

    def test_pos_tags(self):
        nltka = NLTK(pos_tag=True)
        expected = ['Facebook_NNP', 'was_VBD', 'fined_VBN', 'by_IN', 'Hewlett_NNP', 'Packard_NNP', 'for_IN', 'spending_VBG',
                    '100€_CD', 'to_TO', 'buy_VB', 'Cristiano_NNP', 'Ronaldo_NNP', 'from_IN', 'Juventus_NNP']
        result = nltka.process("Facebook was fined by Hewlett Packard for spending 100€ to buy Cristiano Ronaldo from "
                               "Juventus")

        self.assertEqual(expected, result)

    def test_multiple_operations(self):
        # Test for lemmatization with multiple whitespaces removal
        nltka = NLTK(strip_multiple_whitespaces=True, lemmatization=True)
        expected = ["The", "strip", "bat", "be", "hang", "on", "their", "foot", "for", "best"]
        result = nltka.process("The   striped  bats    are    hanging   on   their    feet   for  best")
        self.assertEqual(expected, result)

        # Test for lemmatization with multiple whitespaces removal and URL tagging
        nltka = NLTK(strip_multiple_whitespaces=True, lemmatization=True, url_tagging=True)
        expected = ["The", "strip", "<URL>", "bat", "<URL>", "be", "<URL>", "hang", "on", "their", "foot", "for",
                    "best", "<URL>"]
        result = nltka.process("The   striped http://facebook.com bats https://github.com   are   http://facebook.com "
                               "hanging   on   their    feet   for  best  http://twitter.it")

        self.assertEqual(expected, result)

        # Test for lemmatization, multiple whitespaces removal, URL tagging and stemming
        nltka = NLTK(lemmatization=True, strip_multiple_whitespaces=True, url_tagging=True, stemming=True)
        expected = ["the", "strip", "<url>", "bat", "<url>", "be", "<url>", "hang", "on", "their", "foot", "for",
                    "best", "<url>"]
        result = nltka.process("The   striped http://facebook.com bats https://github.com   are   http://facebook.com "
                               "hanging   on   their    feet   for  best  http://twitter.it")
        self.assertEqual(expected, result)

        # Test for lemmatization, multiple whitespaces removal, URL tagging, stemming, stop words removal,
        # remove punctuation
        nltka = NLTK(lemmatization=True, strip_multiple_whitespaces=True, url_tagging=True, stemming=True,
                     stopwords_removal=True, remove_punctuation=True)
        expected = ["strip", "<url>", "bat", "<url>", "<url>", "hang", "foot", "best", "<url>"]
        result = nltka.process(
            "The   striped, http://facebook.com bats https://github.com   are.   http://facebook.com hanging   On   "
            "their.    feet;   for:  best  http://twitter.it")

        self.assertEqual(expected, result)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Union, List, Tuple, TYPE_CHECKING
import numpy as np
import json

from scipy import sparse

if TYPE_CHECKING:
    from clayrs.content_analyzer.memory_interfaces.memory_interfaces import InformationInterface

from clayrs.content_analyzer.content_representation.representation_container import RepresentationContainer


class FieldRepresentation(ABC):
    """
    Abstract class that generalizes the concept of "field representation",
    a field representation is a semantic way to represent a field of an item.
    """
    __slots__ = ()

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return str(self)

    @property
    @abstractmethod
    def value(self):
        raise NotImplementedError

    def to_json(self):
        """
        The Json representation of each complex representation it's just its string representation, but there might be
        the case in which the representation is more complex than that
        """
        return str(self)


class FeaturesBagField(FieldRepresentation):
    """
    Class for field representation using a bag of features.
    This class can also be used to represent a bag of words: <keyword, score>;
    this representation is produced by the EntityLinking and tf-idf techniques

    Args:
        features (dict<str, object>): the dictionary where features are stored
    """
    __slots__ = ('__scores', '__pos_feature_tuples')

    def __init__(self, sparse_scores: sparse.csc_matrix, pos_feature_tuples: List[Tuple[int, str]]):
        self.__scores = sparse_scores
        self.__pos_feature_tuples = pos_feature_tuples

    @property
    def value(self) -> sparse.csc_matrix:
        """
        Get the features dict

        Returns:
            features (dict<str, object>): the features dict
        """
        return self.__scores

    def to_json(self):
        tuple_representation = np.array([(coordinates_tuple, self.value[coordinates_tuple])
                                         for coordinates_tuple in zip(*self.value.nonzero())], dtype=object)

        return dict(sparse_tfidf=np.array2string(tuple_representation, threshold=np.inf, separator=','),
                    pos_word_tuples=str(self.__pos_feature_tuples),
                    len_vocabulary=self.__scores.shape[1])

    def __str__(self):
        return str(self.__scores)

    def __eq__(self, other):
        return np.array_equal(self.__scores, other.__scores) and self.__pos_feature_tuples == other.__pos_feature_tuples


class SimpleField(FieldRepresentation):
    """
    Class for field with no complex representation.

    Args:
        value (str): string representing the value of the field
    """
    __slots__ = ('__value',)

    def __init__(self, value: object = None):
        self.__value: object = value

    @property
    def value(self) -> object:
        return self.__value

    def __str__(self):
        return str(self.__value)

    def __repr__(self):
        return '{} - {}'.format(str(self), type(self.value))

    def __eq__(self, other):
        return self.__value == other.__value


class EmbeddingField(FieldRepresentation):
    """
    Class for field representation using embeddings (dense numeric vectors)
    this representation is produced by the EmbeddingTechnique.

    Examples:
        shape (4) = [x,x,x,x]
        shape (2,2) = [[x,x],
                       [x,x]]

    Args:
        embedding_array (np.ndarray): embeddings array,
            it can be of different shapes according to the granularity of the technique
    """
    __slots__ = ('__dense_array',)

    def __init__(self, embedding_array: np.ndarray):
        self.__dense_array = embedding_array

    @property
    def value(self) -> np.ndarray:
        return self.__dense_array

    def __str__(self):
        return np.array2string(self.__dense_array, threshold=np.inf, separator=',')

    def __eq__(self, other):
        return self.__dense_array == other.__sparse_array


class IndexField(FieldRepresentation):
    """
    Class for field representation using an index.
    Allows to dynamically retrieve the contents representations serialized in an index.

    Args:
        field_name (str): field's field_name located in the index
            N.B. : it might differ from the original field_name, for example "Plot" might be "Plot_0"
        index_id (int): position of the content in the index
        index (InformationInterface): index from which the data will be retrieved
    """
    __slots__ = ('__field_name', '__index_id', '__index')

    def __init__(self, field_name: str, index_id: int, index: InformationInterface):
        self.__field_name = field_name
        self.__index_id = index_id
        self.__index = index

    @property
    def value(self) -> str:
        return self.__index.get_field(self.__field_name, self.__index_id)

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        return self.__field_name == other.__field_name \
               and self.__index_id == other.__content_id \
               and self.__index == other.__index


class ExogenousPropertiesRepresentation(ABC):
    """
    Output of LodPropertiesRetrieval, different representations exist according to different techniques
    """
    __slots__ = ()

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    def __repr__(self):
        return str(self)

    @property
    @abstractmethod
    def value(self):
        raise NotImplementedError

    def to_json(self):
        """
        The Json representation of each complex representation it's just its string representation, but there might be
        the case in which the representation is more complex than that
        """
        return str(self)


class PropertiesDict(ExogenousPropertiesRepresentation):
    """
    Couples <property name, property value> retrieved by DBPediaMappingTechnique

    Args:
        features: properties in the specified format
    """
    __slots__ = ('__features',)

    def __init__(self, features: Dict[str, str] = None):
        if features is None:
            features = {}

        self.__features: Dict[str, str] = features

    @property
    def value(self):
        """
        Returns: features dictionary
        """
        return self.__features

    def __str__(self):
        return str(self.__features)


class EntitiesProp(ExogenousPropertiesRepresentation):
    """
    Couples <property name, property value> retrieved by DBPediaMappingTechnique

    Args:
        features: properties in the specified format
    """
    __slots__ = ('__features',)

    def __init__(self, features: Dict[str, Dict] = None):
        if features is None:
            features = {}

        self.__features: Dict[str, Dict] = features

    @property
    def value(self):
        """
        Returns: features dictionary
        """
        return self.__features

    def __str__(self):
        return str(self.__features)


class Content:
    """
    Class that represents a content. A content can be an item or a user.
    A content is identified by a string id and is composed by different fields and different exogenous representations.

    The fields and the exogenous representations are stored as Representation Containers and they should not be
    available to the end user. Methods that return Representation Containers must be exclusively used for internal
    purposes.

    Args:
        content_id (str): string identifier
        field_dict (dict[str, RepresentationContainer]): dictionary containing the fields instances for the content,
            and their name as dictionary key
        exogenous_rep_container (RepresentationContainer): different representations of content obtained
            using ExogenousPropertiesRetrieval
    """
    __slots__ = ('__content_id', '__field_dict', '__exogenous_rep_container')

    def __init__(self, content_id: str,
                 field_dict: Dict[str, RepresentationContainer] = None,
                 exogenous_rep_container: RepresentationContainer = None):
        if field_dict is None:
            field_dict = {}  # list o dict
        if exogenous_rep_container is None:
            exogenous_rep_container = RepresentationContainer()

        self.__content_id: str = content_id
        self.__field_dict: Dict[str, RepresentationContainer] = field_dict
        self.__exogenous_rep_container: RepresentationContainer = exogenous_rep_container

    @property
    def content_id(self):
        """
        Getter for the content id
        """
        return self.__content_id

    @property
    def field_dict(self):
        """
        Getter for the dictionary containing the content's fields
        """
        return self.__field_dict

    @property
    def exogenous_rep_container(self):
        """
        Getter for the dictionary containing the content's exogenous representations
        """
        return self.__exogenous_rep_container

    def append_field(self, field_name: str, field: RepresentationContainer):
        """
        Sets a specific field for the content

        Args:
            field_name (str): field name to set
            field (ContentField): represents the data in the field and it will be set for the said field_name
        """
        self.__field_dict[field_name] = field

    def get_field(self, field_name: str) -> RepresentationContainer:
        """
        Getter for the ContentField of a specific field_name

        Args:
            field_name (str): field for which the ContentField will be retrieved
        """
        return self.__field_dict[field_name]

    def remove_field(self, field_name: str):
        """
        Removes the field named field_name from the field dictionary

        Args:
            field_name (str): the name of the field to remove
        """
        self.__field_dict.pop(field_name)

    def append_field_representation(self, field_name: str,
                                    representation: Union[List[FieldRepresentation], FieldRepresentation],
                                    representation_id: Union[List[str], str] = None):
        """
        Adds the given representation(s) to the specific field_name with an external_id corresponding to
        the representation id(s) (if passed). If the field name is not in the content, it will be added.

        It can be passed a single representation and a single id, or a list of representations and a corresponding
        list of ids. In this case the two lists must be of the same length and ids are linked to representations
        by position.
        EXAMPLE:
            representation = [rep1, rep2, rep3]
            representation_id = [id_rep1, id_rep2, id_rep3]

        Args:
            field_name (str): field_name to which the specific representation will be added
            representation (Union[List[FieldRepresentation], FieldRepresentation]): single representation or a list of
                representations for the content
            representation_id (Union[List[str], str]): single name of the field representation or a list of names for
                the field representations.
        """
        if isinstance(representation, list) and representation_id is None:
            representation_id = [None for _ in range(len(representation))]

        if field_name not in self.__field_dict.keys():
            self.__field_dict[field_name] = RepresentationContainer()
        self.__field_dict[field_name].append(representation, representation_id)

    def get_field_representation(self, field_name: str, representation_id: Union[int, str]) -> FieldRepresentation:
        """
        Getter for the FieldRepresentation instance of a specific field representation name

        Args:
            field_name (str): field_name from which the specific representation will be extracted
            representation_id (Union[int, str]): id of the specific representation (either the internal or external id)

        Returns:
            FieldRepresentation: instance of the representation for the name passed as argument
        """
        return self.__field_dict[field_name][representation_id]

    def remove_field_representation(self, field_name: str, representation_id: Union[int, str]):
        """
        Removes a specific representation, identified by the id passed as argument (either internal or external id),
        from a specific field, identified by the field name passed as argument

        Args:
            field_name (str): field_name from which the specific representation will be removed
            representation_id (Union[int, str]): id of the specific representation (either the internal or external id)
        """
        self.__field_dict[field_name].pop(representation_id)

    def append_exogenous_representation(self, exogenous_properties: Union[List[ExogenousPropertiesRepresentation],
                                                                          ExogenousPropertiesRepresentation],
                                        exo_name: Union[List[str], str] = None):
        """
        Sets a specific or multiple exogenous representations for the content. If an exo_name or a list of exo name is
        defined, it will be added as external_id to the representation(s)

        It can be passed a single representation and a single exo_name, or a list of representations and a corresponding
        list of exo_names. In this case the two lists must be of the same length and exo_names are linked to
        representations by position.
        EXAMPLE:
            exogenous_properties = [rep1, rep2, rep3]
           exo_name = [exo_name_rep1, exo_name_rep2, exo_name_rep3]

        Args:
            exogenous_properties (Union[List[ExogenousPropertiesRepresentation], ExogenousPropertiesRepresentation]):
                single exogenous representation or a list of exogenous representations for the content
            exo_name (Union[List[str], str]): single name of the exogenous representation or a list of names for
                the exogenous representations.
        """
        if isinstance(exogenous_properties, list) and exo_name is None:
            exo_name = [None for _ in range(len(exogenous_properties))]

        self.__exogenous_rep_container.append(exogenous_properties, exo_name)

    def get_exogenous_representation(self, exo_name: Union[int, str]) -> ExogenousPropertiesRepresentation:
        """
        Getter for the ExogenousPropertiesRepresentation of a specific exogenous representation name

        Args:
            exo_name (str): representation's name for which the ExogenousPropertiesRepresentation will be retrieved
        """
        return self.__exogenous_rep_container[exo_name]

    def remove_exogenous_representation(self, exo_name: Union[str, int]):
        """
        Removes the exogenous representation named exo_name from the exogenous representation container

        Args:
            exo_name (str): the name of the exogenous representation to remove
        """
        self.__exogenous_rep_container.pop(exo_name)

    def __hash__(self):
        return hash(self.__content_id)

    def __repr__(self):
        return str(self.__content_id)

    def __str__(self):
        content_string = "Content: %s\n" % self.__content_id
        exo_string = "Exogenous representations:\n"
        if len(self.__exogenous_rep_container) != 0:
            exo_string += "%s\n" % self.__exogenous_rep_container
        else:
            exo_string += "\nNo representation found for the Content!\n"

        if len(self.__field_dict.items()) != 0:
            field_string = ''
            for field, rep in self.__field_dict.items():
                field_string += "\nField: %s \n%s\n" % (field, rep)
        else:
            field_string = "Field representations:\n\nNo representation found for the Content fields!"

        return "%s\n%s\n%s\n##############################" % (content_string, exo_string, field_string)

    def __eq__(self, other):
        result = False
        try:
            result = self.__content_id == other.__content_id and self.__field_dict == other.__field_dict
        except AttributeError:
            pass

        return result


class ContentEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Content):
            content = {'content_id': obj.content_id}
            for row in obj.exogenous_rep_container:
                key = f"Exo#{row['internal_id']}"
                # polymorphic call to 'to_json()' function
                content[key] = row['representation'].to_json()

            for field in obj.field_dict:
                field_container = obj.field_dict[field]

                for row in field_container:
                    key = f"{field}#{row['internal_id']}"
                    # polymorphic call to 'to_json()' function
                    content[key] = row['representation'].to_json()

            return content
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)



import time
from unittest import TestCase

import tqdm

from clayrs.utils.context_managers import get_progbar, get_iterator_parallel


class TestContextManagers(TestCase):

    def test_get_progbar(self):

        with get_progbar(range(50), total=50) as pbar:
            self.assertIsInstance(pbar, tqdm.tqdm)

            expected_bar_format = "{desc} {percentage:.0f}%|{bar}| {n:}/{total_fmt} [{elapsed}<{remaining}]"
            result_bar_format = pbar.bar_format

            expected_list = list(range(50))
            result_list = list(pbar)

        self.assertEqual(expected_bar_format, result_bar_format)
        self.assertEqual(expected_list, result_list)

    def test_get_iterator_parallel(self):

        def f(x):
            time.sleep(1)

            return x

        expected_list = list(range(5))

        # single cpu
        with get_iterator_parallel(1, f, list(range(5))) as it:
            result_list = list(it)

        self.assertEqual(expected_list, result_list)

        # multi cpu
        with get_iterator_parallel(2, f, list(range(5))) as it:
            result_list = list(it)

        self.assertEqual(expected_list, result_list)

        # multi cpu with progbar
        with get_iterator_parallel(2, f, list(range(5)), progress_bar=True, total=5) as pbar:

            self.assertIsInstance(pbar, tqdm.tqdm)

            result_list = list(pbar)

        self.assertEqual(expected_list, result_list)



import os
from unittest import TestCase
from clayrs.utils.load_content import load_content_instance
from test import dir_test_files

movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')


class Test(TestCase):
    def test_load_content_instance(self):
        self.assertIsNone(load_content_instance("not_existent", "invalid_item"))
        self.assertIsNotNone(load_content_instance(movies_dir, "tt0112281"))



from random import random
from unittest import TestCase, mock
import numpy as np

from clayrs.content_analyzer.embeddings import Sbert

result_matrix = {
    'this is a phrase': np.array([random() for _ in range(768)]),
    'this is another phrase': np.array([random() for _ in range(768)])
}


def encode(sentence, show_progress_bar):
    return result_matrix[sentence]


class TestSbert(TestCase):

    @mock.patch('clayrs.content_analyzer.embeddings.sbert.SentenceTransformer')
    def test_sbert(self, mocked_model):
        instance = mocked_model.return_value
        instance.get_sentence_embedding_dimension.return_value = 768
        instance.encode.side_effect = encode

        source = Sbert()

        vector_size = source.get_vector_size()

        result = source.load(["this is a phrase", "this is another phrase"])

        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), vector_size)
        self.assertEqual(len(result[1]), vector_size)



from __future__ import annotations
from typing import List, Set, TYPE_CHECKING

import pandas as pd
import numpy as np
from sklearn.metrics import ndcg_score

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction
    from clayrs.recsys.partitioning import Split

from clayrs.evaluation.metrics.metrics import Metric


class RankingMetric(Metric):
    """
    Abstract class that generalize ranking metrics.
    A ranking metric evaluates the quality of the recommendation list
    """
    pass


class NDCG(RankingMetric):
    r"""
    The NDCG (Normalized Discounted Cumulative Gain) metric is calculated for the **single user** by using the sklearn
    implementation, so be sure to check its [documentation][sklearn_link].

    [sklearn_link]: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.ndcg_score.html

    The NDCG of the **entire system** is calculated instead as such:

    $$
    NDCG_{sys} = \frac{\sum_{u} NDCG_u}{|U|}
    $$

    Where:

    - $NDCG_u$ is the NDCG calculated for user :math:`u`
    - $U$ is the set of all users

    The system average excludes NaN values.
    """

    def __str__(self):
        return "NDCG"

    def __repr__(self):
        return "NDCG()"

    def _calc_ndcg(self, ideal_rank: np.array, actual_rank: np.array):
        """
        Private method which calculates the NDCG for a single user using sklearn implementation
        """
        return ndcg_score(ideal_rank, actual_rank)

    def _get_ideal_actual_rank(self, user_predictions: List[Interaction], user_truth: List[Interaction]):
        """
        Private method which calculates two lists, actual_rank list and ideal_rank list.

        actual_rank - given the ranking of the user, for every item 'i' in the ranking it extracts the rating
        that the user has effectively given to 'i' and adds it to the actual_rank list.
        If the item is not present in the truth, a 0 is added to the list.

        ideal_rank - it's the actual_rank list ordered from the highest rating to the lowest one. It represents the
        perfect ranking for the user

        Args:
            user_predictions: list of Interactions object of the recommendation list for the user
            user_truth: list of Interactions object of the truth set for the user
        """
        # important that predicted items is a list, we must maintain the order
        predicted_items = [interaction.item_id for interaction in user_predictions]
        item_score_truth = {interaction.item_id: interaction.score for interaction in user_truth}

        actual_rank = [item_score_truth.get(item_id)
                       if item_score_truth.get(item_id) is not None
                       else 0
                       for item_id in predicted_items]

        ideal_rank = sorted(actual_rank, reverse=True)
        return ideal_rank, actual_rank

    def perform(self, split: Split):

        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], str(self): []}

        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            ideal, actual = self._get_ideal_actual_rank(user_predictions, user_truth)

            if len(ideal) == 1:
                user_ndcg = 1
            else:
                ideal_rank = np.array([ideal])
                actual_rank = np.array([actual])
                user_ndcg = self._calc_ndcg(ideal_rank, actual_rank)

            split_result['user_id'].append(user)
            split_result[str(self)].append(user_ndcg)

        split_result['user_id'].append('sys')
        split_result[str(self)].append(np.nanmean(split_result[str(self)]))

        return pd.DataFrame(split_result)


class NDCGAtK(NDCG):
    r"""
    The NDCG@K (Normalized Discounted Cumulative Gain at K) metric is calculated for the **single user** by using the
    sklearn implementation, so be sure to check its [documentation][sklearn_link].

    [sklearn_link]: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.ndcg_score.html

    The NDCG@K of the **entire system** is calculated instead as such:

    $$
    NDCG@K_{sys} = \frac{\sum_{u} NDCG@K_u}{|U|}
    $$

    Where:

    - $NDCG@K_u$ is the NDCG@K calculated for user $u$
    - $U$ is the set of all users

    The system average excludes NaN values.

    Args:
        k (int): the cutoff parameter
    """

    def __init__(self, k: int):
        self.__k = k

    def __str__(self):
        return "NDCG@{}".format(self.__k)

    def __repr__(self):
        return f'NDCGAtK(k={self.__k})'

    def _calc_ndcg(self, ideal_rank: np.array, actual_rank: np.array):
        return ndcg_score(ideal_rank, actual_rank, k=self.__k)


class MRR(RankingMetric):
    r"""
    The MRR (Mean Reciprocal Rank) metric is a system wide metric, so only its result it will be returned and not those
    of every user.
    MRR is calculated as such:

    $$
    MRR_{sys} = \frac{1}{|Q|}\cdot\sum_{i=1}^{|Q|}\frac{1}{rank(i)}
    $$

    Where:

    - $Q$ is the set of recommendation lists
    - $rank(i)$ is the position of the first relevant item in the i-th recommendation list

    The MRR metric needs to discern relevant items from the not relevant ones: in order to do that, one could pass a
    custom `relevant_threshold` parameter that will be applied to every user, so that if a rating of an item
    is >= relevant_threshold, then it's relevant, otherwise it's not.
    If no `relevant_threshold` parameter is passed then, for every user, its mean rating score will be used

    Args:
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
    """

    def __init__(self, relevant_threshold: float = None):
        self.__relevant_threshold = relevant_threshold

    @property
    def relevant_threshold(self):
        return self.__relevant_threshold

    def __str__(self):
        return "MRR"

    def __repr__(self):
        return f'MRR(relevant_threshold={self.relevant_threshold})'

    def calc_reciprocal_rank(self, user_predictions: List[Interaction], user_truth_relevant_items: Set[Interaction]):
        """
        Method which calculates the RR (Reciprocal Rank) for a single user

        Args:
            user_predictions: list of Interactions object of the recommendation list for the user
            user_truth_relevant_items: list of relevant Interactions object of the truth set for the user
        """

        reciprocal_rank = 0
        i = 1
        for interaction_pred in user_predictions:
            if interaction_pred.item_id in user_truth_relevant_items:
                reciprocal_rank = 1 / i
                break  # We only need the first relevant item position in the rank

            i += 1

        return reciprocal_rank

    def perform(self, split: Split) -> pd.DataFrame:
        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], str(self): []}

        rr_list = []
        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            relevant_threshold = self.relevant_threshold
            if relevant_threshold is None:
                relevant_threshold = np.nanmean([interaction.score for interaction in user_truth])

            user_truth_relevant_items = set(interaction.item_id for interaction in user_truth
                                            if interaction.score >= relevant_threshold)

            if len(user_truth_relevant_items) != 0:
                user_reciprocal_rank = self.calc_reciprocal_rank(user_predictions, user_truth_relevant_items)
            else:
                user_reciprocal_rank = np.nan

            rr_list.append(user_reciprocal_rank)

        # trick to check for nan values, if all values are nan then an exception is thrown
        if all(rr != rr for rr in rr_list):
            raise ValueError("No user has a rating above the given threshold! Try lower it")

        mrr = np.nanmean(rr_list)

        split_result['user_id'].append('sys')
        split_result[str(self)].append(mrr)

        return pd.DataFrame(split_result)


class MRRAtK(MRR):
    r"""
    The MRR@K (Mean Reciprocal Rank at K) metric is a system wide metric, so only its result will be returned and
    not those of every user.
    MRR@K is calculated as such

    $$
    MRR@K_{sys} = \frac{1}{|Q|}\cdot\sum_{i=1}^{K}\frac{1}{rank(i)}
    $$

    Where:

    - $K$ is the cutoff parameter
    - $Q$ is the set of recommendation lists
    - $rank(i)$ is the position of the first relevant item in the i-th recommendation list


    Args:
        k (int): the cutoff parameter. It must be >= 1, otherwise a ValueError exception is raised
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used

    Raises:
        ValueError: if an invalid cutoff parameter is passed (0 or negative)
    """

    def __init__(self, k: int, relevant_threshold: float = None):
        if k < 1:
            raise ValueError('k={} not valid! k must be >= 1!'.format(k))
        self.__k = k
        super().__init__(relevant_threshold)

    @property
    def k(self):
        return self.__k

    def __str__(self):
        return "MRR@{}".format(self.k)

    def __repr__(self):
        return f'MRRAtK(relevant_threshold={self.relevant_threshold})'

    def calc_reciprocal_rank(self, user_predictions: List[Interaction], user_truth_relevant_items: Set[Interaction]):
        """
        Method which calculates the RR (Reciprocal Rank) for a single user

        Args:
            user_predictions: list of Interactions object of the recommendation list for the user
            user_truth_relevant_items: list of relevant Interactions object of the truth set for the user
        """
        user_predictions_cut = user_predictions[:self.k]

        return super().calc_reciprocal_rank(user_predictions_cut, user_truth_relevant_items)


class MAP(RankingMetric):
    r"""

    The $MAP$ metric (*Mean average Precision*) is a ranking metric computed by first calculating the $AP$
    (*Average Precision*) for each user and then taking its mean.

    The $AP$ is calculated as such for the single user:

    $$
    AP_u = \frac{1}{m_u}\sum_{i=1}^{N_u}P(i)\cdot rel(i)
    $$

    Where:

    - $m_u$ is the number of relevant items for the user $u$
    - $N_u$ is the number of recommended items for the user $u$
    - $P(i)$ is the precision computed at cutoff $i$
    - $rel(i)$ is an indicator variable that says whether the i-th item is relevant ($rel(i)=1$) or not ($rel(i)=0$)

    After computing the $AP$ for each user, we can compute the $MAP$ for the whole system:

    $$
    MAP_{sys} = \frac{1}{|U|}\sum_{u}AP_u
    $$

    This metric will return the $AP$ computed for each user in the dataframe containing users results, and the $MAP$
    computed for the whole system in the dataframe containing system results

    Args:
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
    """
    def __init__(self, relevant_threshold: float = None):
        self.relevant_threshold = relevant_threshold

    def _compute_ap(self, user_predictions: List[Interaction], user_truth_relevant_items: Set[str]):
        tp = 0
        cumulative_precision = 0
        for position, interaction in enumerate(user_predictions, start=1):

            # this 'if' acts as the relevance indicator in the AP formula
            if interaction.item_id in user_truth_relevant_items:
                tp += 1
                cumulative_precision += tp / position

        user_ap = (1 / len(user_truth_relevant_items)) * cumulative_precision

        return user_ap

    def perform(self, split: Split):
        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], 'AP': []}

        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            relevant_threshold = self.relevant_threshold
            if relevant_threshold is None:
                relevant_threshold = np.nanmean([interaction.score for interaction in user_truth])

            user_truth_relevant_items = set(interaction.item_id for interaction in user_truth
                                            if interaction.score >= relevant_threshold)

            if len(user_truth_relevant_items) != 0:
                user_ap = self._compute_ap(user_predictions, user_truth_relevant_items)
            else:
                user_ap = np.nan

            split_result['user_id'].append(user)
            split_result['AP'].append(user_ap)  # for users we are computing Average Precision

        df_users = pd.DataFrame(split_result)

        # for the system we are computing Mean Average Precision
        sys_map = np.nanmean(df_users['AP'])
        df_sys = pd.DataFrame({'user_id': ['sys'], str(self): [sys_map]})

        df = pd.concat([df_users, df_sys])

        return df

    def __str__(self):
        return "MAP"


class MAPAtK(MAP):
    r"""

    The $MAP@K$ metric (*Mean average Precision At K*) is a ranking metric computed by first calculating the $AP@K$
    (*Average Precision At K*) for each user and then taking its mean.

    The $AP@K$ is calculated as such for the single user:

    $$
    AP@K_u = \frac{1}{m_u}\sum_{i=1}^{K}P(i)\cdot rel(i)
    $$

    Where:

    - $m_u$ is the number of relevant items for the user $u$
    - $K$ is the cutoff value
    - $P(i)$ is the precision computed at cutoff $i$
    - $rel(i)$ is an indicator variable that says whether the i-th item is relevant ($rel(i)=1$) or not ($rel(i)=0$)

    After computing the $AP@K$ for each user, we can compute the $MAP@K$ for the whole system:

    $$
    MAP@K_{sys} = \frac{1}{|U|}\sum_{u}AP@K_u
    $$

    This metric will return the $AP@K$ computed for each user in the dataframe containing users results, and the $MAP@K$
    computed for the whole system in the dataframe containing system results

    Args:
        k (int): the cutoff parameter. It must be >= 1, otherwise a ValueError exception is raised
        relevant_threshold (float): parameter needed to discern relevant items and non-relevant items for every
            user. If not specified, the mean rating score of every user will be used
    """
    def __init__(self, k: int, relevant_threshold: float = None):
        super().__init__(relevant_threshold)
        self.k = k

    def _compute_ap(self, user_predictions: List[Interaction], user_truth_relevant_items: Set[str]):
        user_predictions = user_predictions[:self.k]

        return super()._compute_ap(user_predictions, user_truth_relevant_items)


class Correlation(RankingMetric):
    r"""
    The Correlation metric calculates the correlation between the ranking of a user and its ideal ranking.
    The currently correlation methods implemented are:

    - `pearson`
    - `kendall`
    - `spearman`

    Every correlation method is implemented by the pandas library, so read its [documentation][pd_link] for more

    [pd_link]: https://pandas.pydata.org/docs/reference/api/pandas.Series.corr.html

    The correlation metric is calculated as such for the **single user**:

    $$
    Corr_u = Corr(ranking_u, ideal\_ranking_u)
    $$

    Where:

    - $ranking_u$ is ranking of the user
    - $ideal\_ranking_u$ is the ideal ranking for the user

    The ideal ranking is calculated based on the rating inside the *ground truth* of the user

    The Correlation metric calculated for the **entire system** is simply the average of every $Corr$:

    $$
    Corr_{sys} = \frac{\sum_{u} Corr_u}{|U|}
    $$

    Where:

    - $Corr_u$ is the correlation of the user $u$
    - $U$ is the set of all users

    The system average excludes NaN values.

    It's also possible to specify a cutoff parameter thanks to the 'top_n' parameter: if specified, only the first
    $n$ results of the recommendation list will be used in order to calculate the correlation

    Args:
        method (str): The correlation method to use. It must be 'pearson', 'kendall' or 'spearman', otherwise a
            ValueError exception is raised. By default is 'pearson'
        top_n (int): Cutoff parameter, if specified only the first n items of the recommendation list will be used
            in order to calculate the correlation

    Raises:
        ValueError: if an invalid method parameter is passed
    """

    def __init__(self, method: str = 'pearson', top_n: int = None):
        valid = {'pearson', 'kendall', 'spearman'}
        self.__method = method.lower()

        if self.__method not in valid:
            raise ValueError("Method {} is not supported! Methods available:\n"
                             "{}".format(method, valid))

        self.__top_n = top_n

    def __str__(self):
        name = self.__method
        if self.__top_n is not None:
            name += " - Top {}".format(self.__top_n)

        return name

    def __repr__(self):
        return f'Correlation(method={self.__method}, top_n={self.__top_n})'

    def perform(self, split: Split) -> pd.DataFrame:
        pred = split.pred
        truth = split.truth

        split_result = {'user_id': [], str(self): []}
        for user in set(truth.user_id_column):
            user_predictions = pred.get_user_interactions(user)
            user_truth = truth.get_user_interactions(user)

            ideal, actual = self._get_ideal_actual_rank(user_predictions, user_truth)

            if len(actual) < 2:
                coef = np.nan
            else:
                ideal_ranking = pd.Series(ideal)
                actual_ranking = pd.Series(actual)
                coef = actual_ranking.corr(ideal_ranking, method=self.__method)

            split_result['user_id'].append(user)
            split_result[str(self)].append(coef)

        split_result['user_id'].append('sys')
        split_result[str(self)].append(np.nanmean(split_result[str(self)]))

        return pd.DataFrame(split_result)

    def _get_ideal_actual_rank(self, user_predictions: List[Interaction], user_truth: List[Interaction]):

        if self.__top_n is not None:
            user_predictions = user_predictions[:self.__top_n]

        # sorting truth on score values
        user_truth_ordered = sorted(user_truth, key=lambda interaction: interaction.score, reverse=True)
        ideal_rank = [interaction.item_id for interaction in user_truth_ordered]

        predicted_items = [interaction.item_id for interaction in user_predictions]

        actual_rank = [predicted_items.index(item)
                       for item in ideal_rank
                       if item in set(predicted_items)]

        # the ideal rank is basically 0, 1, 2, 3 etc.
        return [i for i in range(len(ideal_rank))], actual_rank



import unittest
import pandas as pd
from sklearn.metrics import ndcg_score
import numpy as np

from clayrs.content_analyzer import Ratings
from clayrs.evaluation.metrics.ranking_metrics import NDCG, Correlation, MRR, NDCGAtK, MRRAtK, MAP, MAPAtK
from clayrs.evaluation.eval_pipeline_modules.metric_evaluator import Split

pred_only_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u2', 'u2'],
     'item_id': ['inew1', 'inew2', 'inew3', 'inew4'],
     'score': [650, 600, 500, 650]})
pred_only_new_items = Ratings.from_dataframe(pred_only_new_items)

pred_w_new_items = pd.DataFrame(
    {'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2', 'u2'],
     'item_id': ['i2', 'i1', 'i4', 'i5', 'i6', 'i3', 'i8', 'i9', 'i4', 'i6', 'i1', 'i8'],
     'score': [650, 600, 500, 400, 300, 220, 100, 50, 350, 200, 100, 50]})
pred_w_new_items = Ratings.from_dataframe(pred_w_new_items)

pred_only_one_item = pd.DataFrame(
    {'user_id': ['u1', 'u2'],
     'item_id': ['i4', 'i8'],
     'score': [650, 600]})
pred_only_one_item = Ratings.from_dataframe(pred_only_one_item)

truth = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                      'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                      'score': [3, 2, 3, 1, 2, 4, 3, 3]})
truth = Ratings.from_dataframe(truth)

split_only_new = Split(pred_only_new_items, truth)
split_w_new_items = Split(pred_w_new_items, truth)
split_only_one = Split(pred_only_one_item, truth)


def for_each_method(test_func):
    def wrapper(self, *args, **kwargs):
        for method in self.methods_list:
            with self.subTest(current_method=method):
                test_func(*((self, method) + args), **kwargs)

    return wrapper


class TestNDCG(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = NDCG()

    def test_perform_only_new(self):
        metric = self.metric

        result_only_new = metric.perform(split_only_new)

        ndcgs_predicted = set(result_only_new[str(metric)])
        ndcgs_expected = {0}
        self.assertEqual(ndcgs_expected, ndcgs_predicted)

    def test_perform_w_new_items(self):
        metric = self.metric

        result_w_new_items = metric.perform(split_w_new_items)

        u1_actual = [[2, 3, 1, 0, 2, 3, 0, 0]]
        u1_ideal = [[3, 3, 2, 2, 1, 0, 0, 0]]

        u1_expected_ndcg = ndcg_score(u1_ideal, u1_actual)
        u1_result_ndcg = float(result_w_new_items.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected_ndcg, u1_result_ndcg)

        u2_actual = [[3, 0, 4, 3]]
        u2_ideal = [[4, 3, 3, 0]]

        u2_expected_ndcg = ndcg_score(u2_ideal, u2_actual)
        u2_result_ndcg = float(result_w_new_items.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected_ndcg, u2_result_ndcg)

        sys_expected_ndcg = (u1_expected_ndcg + u2_expected_ndcg) / 2
        sys_result_ndcg = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_ndcg, sys_result_ndcg)

    def test_perform_only_one(self):
        metric = self.metric

        result_only_one = metric.perform(split_only_one)

        ndcgs_predicted = set(result_only_one[str(metric)])
        ndcgs_expected = {1}

        self.assertEqual(ndcgs_expected, ndcgs_predicted)


class TestNDCGAtK(unittest.TestCase):

    def test_perform(self):
        k = 2
        metric = NDCGAtK(k=k)

        result_w_new_items = metric.perform(split_w_new_items)

        u1_actual = [[2, 3, 1, 0, 2, 3, 0, 0]]
        u1_ideal = [[3, 3, 2, 2, 1, 0, 0, 0]]

        u1_expected_ndcg = ndcg_score(u1_ideal, u1_actual, k=k)
        u1_result_ndcg = float(result_w_new_items.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected_ndcg, u1_result_ndcg)

        u2_actual = [[3, 0, 4, 3]]
        u2_ideal = [[4, 3, 3, 0]]

        u2_expected_ndcg = ndcg_score(u2_ideal, u2_actual, k=k)
        u2_result_ndcg = float(result_w_new_items.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected_ndcg, u2_result_ndcg)

        sys_expected_ndcg = (u1_expected_ndcg + u2_expected_ndcg) / 2
        sys_result_ndcg = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_ndcg, sys_result_ndcg)

    def test_perform_w_greater_k(self):
        k = 99
        metric = NDCGAtK(k=k)

        result = metric.perform(split_w_new_items)

        expected = NDCG().perform(split_w_new_items)

        result = np.sort(result, axis=0)
        expected = np.sort(expected, axis=0)

        self.assertTrue(np.array_equal(expected, result))


class TestMRR(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.metric = MRR(relevant_threshold=3)

        cls.metric_mean_threshold = MRR()
        # mean u1 = 2.2; mean u2 = 3.33

    def test_perform_only_new(self):
        metric = self.metric

        result_only_new = metric.perform(split_only_new)

        mrr_expected = 0
        mrr_predicted = float(result_only_new[str(metric)])

        self.assertEqual(mrr_expected, mrr_predicted)

        # u1 = [0, 0], u2 = [0, 0]

    def test_perform_only_new_mean(self):
        metric = self.metric_mean_threshold

        result_only_new = metric.perform(split_only_new)

        mrr_expected = 0
        mrr_predicted = float(result_only_new[str(metric)])

        self.assertEqual(mrr_expected, mrr_predicted)

        # u1 = [0, 0], u2 = [0, 0]

    def test_perform_w_new_items(self):
        metric = self.metric

        result_w_new_items = metric.perform(split_w_new_items)

        u1_expected_rr = 1 / 2
        u2_expected_rr = 1

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2

        sys_result_mrr = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_mrr, sys_result_mrr)

        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [1, 0, 1, 1]

    def test_perform_w_new_items_mean(self):
        metric = self.metric_mean_threshold

        result_w_new_items = metric.perform(split_w_new_items)

        u1_expected_rr = 1 / 2
        u2_expected_rr = 1 / 3

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2

        sys_result_mrr = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_mrr, sys_result_mrr)

        # u1 = [0, 1, 0, 0, 0, 1, 0, 0], u2 = [0, 0, 1, 0]

    def test_perform_only_one(self):
        metric = self.metric

        result_only_one = metric.perform(split_only_one)

        u1_expected_rr = 0
        u2_expected_rr = 1 / 1

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2
        sys_result_mrr = float(result_only_one[str(metric)])

        self.assertEqual(sys_expected_mrr, sys_result_mrr)

        # u1 = [0], u2 = [1]

    def test_perform_only_one_mean(self):
        metric = self.metric_mean_threshold

        result_only_one = metric.perform(split_only_one)

        u1_expected_rr = 0
        u2_expected_rr = 0

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2
        sys_result_mrr = float(result_only_one[str(metric)])

        self.assertEqual(sys_expected_mrr, sys_result_mrr)

        # u1 = [0], u2 = [0]

    def test_perform_no_relevant_u1(self):
        metric_no_relevant_u1 = MRR(relevant_threshold=4)

        result = metric_no_relevant_u1.perform(split_w_new_items)

        expected_sys = 1 / 3  # only u2 result will be considered
        result_sys = float(result.query('user_id == "sys"')[str(metric_no_relevant_u1)])
        self.assertAlmostEqual(expected_sys, result_sys)

        # u1 has no relevant items, u2 = [0, 0, 1, 0]

    def test_no_relevant_items(self):
        truth_no_rel = pd.DataFrame({'user_id': ['u1', 'u1', 'u1', 'u1', 'u1', 'u2', 'u2', 'u2'],
                                     'item_id': ['i1', 'i2', 'i3', 'i4', 'i6', 'i1', 'i8', 'i4'],
                                     'score': [3, 2, 3, 1, 2, 2, 3, 3]})
        truth_no_rel = Ratings.from_dataframe(truth_no_rel)

        split_no_rel = Split(pred_w_new_items, truth_no_rel)

        # any ClassificationMetric will work
        metric = MRR(relevant_threshold=4)

        with self.assertRaises(ValueError):
            metric.perform(split_no_rel)


class TestMRRAtK(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        k = 2
        cls.k = k

        cls.metric_mean_threshold = MRRAtK(k)

    def test_perform(self):
        metric = MRRAtK(self.k, relevant_threshold=3)

        result_w_new_items = metric.perform(split_w_new_items)

        u1_expected_rr = 1 / 2
        u2_expected_rr = 1

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2

        sys_result_mrr = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_mrr, sys_result_mrr)

        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@k = [0, 1], u2@k = [1, 0]

    def test_perform_w_greater_k(self):
        metric = MRRAtK(99, relevant_threshold=3)

        result = metric.perform(split_w_new_items)
        expected = MRR(relevant_threshold=3).perform(split_w_new_items)

        result = np.sort(result, axis=0)
        expected = np.sort(expected, axis=0)

        self.assertTrue(np.array_equal(expected, result))

    def test_perform_w_new_items_mean(self):
        metric = MRRAtK(self.k)

        result_w_new_items = metric.perform(split_w_new_items)

        u1_expected_rr = 1 / 2
        u2_expected_rr = 0

        sys_expected_mrr = (u1_expected_rr + u2_expected_rr) / 2

        sys_result_mrr = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_mrr, sys_result_mrr)

        # mean u1 = 2.2; mean u2 = 3.33
        # u1_total = [0, 1, 0, 0, 0, 1, 0, 0], u2_total = [1, 0, 1, 1]
        # u1@k = [0, 1], u2@k = [0, 0]

    def test_k_not_valid(self):
        with self.assertRaises(ValueError):
            MRRAtK(k=-2)
            MRRAtK(k=0)


class TestMAP(unittest.TestCase):

    def test_perform_w_new(self):
        split = Split(pred_w_new_items, truth)

        metric = MAP(relevant_threshold=3)
        df_result = metric.perform(split)

        u1_result = float(df_result.query('user_id == "u1"')['AP'])
        u1_expected = 1/2 * (1/2 + 2/6)
        self.assertAlmostEqual(u1_expected, u1_result)

        u2_result = float(df_result.query('user_id == "u2"')['AP'])
        u2_expected = 1/3 * (1/1 + 2/3 + 3/4)
        self.assertAlmostEqual(u2_expected, u2_result)

        sys_result = float(df_result.query('user_id == "sys"')['MAP'])
        sys_expected = (u1_expected + u2_expected) / 2
        self.assertAlmostEqual(sys_expected, sys_result)

        # WITH MEAN as relevant threshold
        metric = MAP(relevant_threshold=None)
        df_result = metric.perform(split)

        u1_result = float(df_result.query('user_id == "u1"')['AP'])
        u1_expected = 1/2 * (1/2 + 2/6)
        self.assertAlmostEqual(u1_expected, u1_result)

        u2_result = float(df_result.query('user_id == "u2"')['AP'])
        u2_expected = 1/1 * (1/3)
        self.assertAlmostEqual(u2_expected, u2_result)

        sys_result = float(df_result.query('user_id == "sys"')['MAP'])
        sys_expected = (u1_expected + u2_expected) / 2
        self.assertAlmostEqual(sys_expected, sys_result)

    def test_perform_only_one(self):
        split = Split(pred_only_one_item, truth)

        metric = MAP(relevant_threshold=3)
        df_result = metric.perform(split)

        u1_result = float(df_result.query('user_id == "u1"')['AP'])
        u1_expected = 0
        self.assertAlmostEqual(u1_expected, u1_result)

        u2_result = float(df_result.query('user_id == "u2"')['AP'])
        u2_expected = 1/3 * (1/1)
        self.assertAlmostEqual(u2_expected, u2_result)

        sys_result = float(df_result.query('user_id == "sys"')['MAP'])
        sys_expected = (u1_expected + u2_expected) / 2
        self.assertAlmostEqual(sys_expected, sys_result)

    def test_perform_only_new(self):

        split = Split(pred_only_new_items, truth)

        metric = MAP(relevant_threshold=2)
        df_result = metric.perform(split)

        u1_result = float(df_result.query('user_id == "u1"')['AP'])
        u1_expected = 0
        self.assertEqual(u1_expected, u1_result)

        u2_result = float(df_result.query('user_id == "u2"')['AP'])
        u2_expected = 0
        self.assertEqual(u2_expected, u2_result)

        sys_result = float(df_result.query('user_id == "sys"')['MAP'])
        sys_expected = 0
        self.assertEqual(sys_expected, sys_result)

    def test_perform_nan(self):
        # RELEVANT THRESHOLD greater than all rating given by u1
        split = Split(pred_w_new_items, truth)
        metric = MAP(relevant_threshold=4)
        df_result = metric.perform(split)

        u1_result = df_result.query('user_id == "u1"')['AP'].values
        self.assertTrue(pd.isna(u1_result))

        u2_result = float(df_result.query('user_id == "u2"')['AP'])
        u2_expected = 1/1 * (1/3)
        self.assertAlmostEqual(u2_expected, u2_result)

        sys_result = float(df_result.query('user_id == "sys"')['MAP'])
        sys_expected = u2_expected  # since u1 is nan only u2 will matter for MAP computation
        self.assertAlmostEqual(sys_expected, sys_result)


class TestMAPAtK(unittest.TestCase):

    def test__compute_ap(self):
        relevant_threshold = 3

        user_predictions = pred_w_new_items.get_user_interactions('u1')
        user_truth = truth.get_user_interactions('u1')
        user_truth_relevant_items = set(interaction.item_id for interaction in user_truth
                                        if interaction.score >= relevant_threshold)

        metric = MAPAtK(k=2)

        result_u1_ap = metric._compute_ap(user_predictions, user_truth_relevant_items)
        expected_u1_ap = 1/2 * 1/2
        self.assertAlmostEqual(expected_u1_ap, result_u1_ap)


class TestCorrelation(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.methods_list = ['pearson', 'spearman', 'kendall']
        cls.top_n = 2

    @for_each_method
    def test_perform_only_new(self, method: str):
        metric = Correlation(method)

        result_only_new = metric.perform(split_only_new)

        pearsons_predicted = result_only_new[str(metric)]

        self.assertTrue(all(pd.isna(pearsons_predicted)))

    @for_each_method
    def test_perform_only_new_top_n(self, method: str):
        metric = Correlation(method, top_n=self.top_n)

        result_only_new = metric.perform(split_only_new)

        pearsons_predicted = result_only_new[str(metric)]

        self.assertTrue(all(pd.isna(pearsons_predicted)))

    @for_each_method
    def test_perform_w_new_items(self, method):
        metric = Correlation(method)

        result_w_new_items = metric.perform(split_w_new_items)

        u1_actual = pd.Series([1, 5, 0, 4, 2])
        u1_ideal = pd.Series([0, 1, 2, 3, 4])

        u1_expected_pearson = u1_actual.corr(u1_ideal, method)
        u1_result_pearson = float(result_w_new_items.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected_pearson, u1_result_pearson)

        u2_actual = pd.Series([2, 3, 0])
        u2_ideal = pd.Series([0, 1, 2])

        u2_expected_pearson = u2_actual.corr(u2_ideal, method)
        u2_result_pearson = float(result_w_new_items.query('user_id == "u2"')[str(metric)])

        self.assertAlmostEqual(u2_expected_pearson, u2_result_pearson)

        sys_expected_pearson = (u1_expected_pearson + u2_expected_pearson) / 2
        sys_result_pearson = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_pearson, sys_result_pearson)

    @for_each_method
    def test_perform_w_new_items_top_n(self, method: str):
        metric = Correlation(method, top_n=self.top_n)

        result_w_new_items = metric.perform(split_w_new_items)

        u1_actual = pd.Series([1, 0])
        u1_ideal = pd.Series([0, 1, 2, 3, 4])

        u1_expected_pearson = u1_actual.corr(u1_ideal, method)
        u1_result_pearson = float(result_w_new_items.query('user_id == "u1"')[str(metric)])

        self.assertAlmostEqual(u1_expected_pearson, u1_result_pearson)

        u2_actual = pd.Series([0])
        u2_ideal = pd.Series([0, 1, 2])

        u2_expected_pearson = u2_actual.corr(u2_ideal, method)
        u2_result_pearson = float(result_w_new_items.query('user_id == "u2"')[str(metric)])

        self.assertTrue(np.isnan(u2_expected_pearson))
        self.assertTrue(np.isnan(u2_result_pearson))

        sys_expected_pearson = u1_expected_pearson  # the mean doesn't consider nan values
        sys_result_pearson = float(result_w_new_items.query('user_id == "sys"')[str(metric)])

        self.assertAlmostEqual(sys_expected_pearson, sys_result_pearson)

    @for_each_method
    def test_perform_only_one(self, method: str):
        metric = Correlation(method)

        result_only_one = metric.perform(split_only_one)

        pearsons_predicted = result_only_one[str(metric)]

        self.assertTrue(all(pd.isna(pearsons_predicted)))

    @for_each_method
    def test_perform_only_one_top_n(self, method: str):
        metric = Correlation(method, top_n=self.top_n)

        result_only_one = metric.perform(split_only_one)

        pearsons_predicted = result_only_one[str(metric)]

        self.assertTrue(all(pd.isna(pearsons_predicted)))


if __name__ == '__main__':
    unittest.main()
